Type: Package
Package: worldmet
Title: Import Surface Meteorological Data from NOAA
Version: 1.1.0
Authors@R: c(
    person("David", "Carslaw", , "david.carslaw@york.ac.uk", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-0991-950X")),
    person("Jack", "Davison", , "jack.davison@wsp.com", role = "aut",
           comment = c(ORCID = "0000-0003-2653-6615"))
  )
Description: Functions to import data from more than 30,000 surface
    meteorological sites around the world managed by the National Oceanic
    and Atmospheric Administration (NOAA) Global Historical Climate
    Network (GHCN) and Integrated Surface Database (ISD).
License: MIT + file LICENSE
URL: https://openair-project.github.io/worldmet/,
        https://github.com/openair-project/worldmet
BugReports: https://github.com/openair-project/worldmet/issues
Depends: R (>= 4.1.0)
Imports: carrier, cli, dplyr, lifecycle, mirai, purrr (>= 1.1.0),
        readr, rlang, sf, tibble, tidyr
Suggests: arrow, knitr, leaflet, rmarkdown
ByteCompile: true
Config/Needs/website: openair-project/openairpkgdown,openair
Encoding: UTF-8
Language: en-GB
LazyData: true
LazyLoad: true
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-18 07:31:49 UTC; JD38
Author: David Carslaw [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-0991-950X>),
  Jack Davison [aut] (ORCID: <https://orcid.org/0000-0003-2653-6615>)
Maintainer: David Carslaw <david.carslaw@york.ac.uk>
Repository: CRAN
Date/Publication: 2026-05-18 08:00:02 UTC
Built: R 4.6.0; ; 2026-05-18 09:27:30 UTC; unix

# Generated by roxygen2: do not edit by hand

export(exportADMS)
export(getMeta)
export(getMetaLive)
export(importNOAA)
export(importNOAAlite)
export(import_ghcn_countries)
export(import_ghcn_daily)
export(import_ghcn_hourly)
export(import_ghcn_inventory)
export(import_ghcn_monthly_prcp)
export(import_ghcn_monthly_temp)
export(import_ghcn_stations)
export(import_isd_hourly)
export(import_isd_lite)
export(import_isd_stations)
export(import_isd_stations_live)
export(write_adms)
export(write_met)
importFrom(carrier,crate)
importFrom(lifecycle,deprecated)
importFrom(mirai,daemons)
importFrom(rlang,"%||%")
importFrom(rlang,.data)
importFrom(tibble,tibble)

<div align="center">

<img src="man/figures/logo.png" height="200"/>

## **worldmet**
### open source tools to access NOAA meteorological observations

<!-- badges: start -->
[![R-CMD-check](https://github.com/openair-project/worldmet/workflows/R-CMD-check/badge.svg)](https://github.com/openair-project/worldmet/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/worldmet)](https://CRAN.R-project.org/package=worldmet)
[![metacran downloads](https://cranlogs.r-pkg.org/badges/grand-total/worldmet)](https://cran.r-project.org/package=worldmet)
<br>
[![github](https://img.shields.io/badge/CODE-github-black?logo=github)](https://github.com/openair-project/worldmet)
[![website](https://img.shields.io/badge/DOCS-website-black)](https://openair-project.github.io/worldmet/)
[![book](https://img.shields.io/badge/DOCS-book-black)](https://openair-project.github.io/book/)
<!-- badges: end -->

</div>

**worldmet** provides an easy way to access data from the [NOAA Global Historical Climate Network](https://www.ncei.noaa.gov/products/global-historical-climatology-network-hourly) and the [NOAA Integrated Surface Database](https://www.ncei.noaa.gov/products/land-based-station/integrated-surface-database) (ISD). The GHCN contains detailed surface meteorological data from around the world for over 35,000 locations. The data available through the package work very well with the [**openair**](https://github.com/openair-project/openair) package.

<div align="center">

*Part of the openair toolkit*

[![openair](https://img.shields.io/badge/openair_core-06D6A0?style=flat-square)](https://openair-project.github.io/openair/) | 
[![worldmet](https://img.shields.io/badge/worldmet-26547C?style=flat-square)](https://openair-project.github.io/worldmet/) | 
[![openairmaps](https://img.shields.io/badge/openairmaps-FFD166?style=flat-square)](https://openair-project.github.io/openairmaps/) | 
[![deweather](https://img.shields.io/badge/deweather-EF476F?style=flat-square)](https://openair-project.github.io/deweather/)

</div>

<hr>

## 💡 Core Features

**worldmet** has a small handful of core functionality.

- **Access metadata** using `import_ghcn_stations()`.

- **Import monitoring data** using `import_ghcn_hourly()`. Data is in a format ready to use with, for example, `openair::windRose()`.

- **Write files in ADMS format** using `write_adms()`.

<div align="center">
<img src="man/figures/feature-banner.png" width="800">
</div>

<hr>

## 📖 Documentation

All **worldmet** functions are fully documented; access documentation using R in your IDE of choice.

```r
?worldmet::import_ghcn_stations
```

Documentation is also hosted online on the **package website**.

[![website](https://img.shields.io/badge/website-documentation-blue)](https://openair-project.github.io/worldmet/)

A guide to the openair toolkit can be found in the **online book**, which contains lots of code snippets, demonstrations of functionality, and ideas for the application of **openair**'s various functions.

[![book](https://img.shields.io/badge/book-code_demos_and_ideas-blue)](https://openair-project.github.io/book/)

<hr>

## 🗃️ Installation

**worldmet** can be installed from **CRAN** with:

``` r
install.packages("worldmet")
```

You can also install the development version of **worldmet** from GitHub using `{pak}`:

``` r
# install.packages("pak")
pak::pak("openair-project/worldmet")
```

<hr>

🏛️ **worldmet** is primarily maintained by [David Carslaw](https://github.com/davidcarslaw).

📃 **worldmet** is licensed under the [MIT License](https://openair-project.github.io/worldmet/LICENSE.html).

🧑‍💻 Contributions are welcome from the wider community. See the [contributing guide](https://openair-project.github.io/worldmet/CONTRIBUTING.html) and [code of conduct](https://openair-project.github.io/worldmet/CODE_OF_CONDUCT.html) for more information.

#' Import station metadata for the Global Historical Climatology Network
#'
#' This function accesses a full list of GHCN stations available through either
#' the GHCNh or GHCNd. As well as the station `id`, needed for importing
#' measurement data, useful geographic and network metadata is also returned.
#'
#' @param name,country,state String values to use to filter the metadata for
#'   specific site names, countries and states. `country` and `state` are
#'   matched exactly to codes accessed using [import_ghcn_countries()]. `name`
#'   is searched as a sub-string case insensitively.
#'
#' @param lat,lng,n_max Decimal latitude (`lat`) and longitude (`lng`) (or other
#'   Y/X coordinate if using a different `crs`). If provided, the `n_max`
#'   closest ISD stations to this coordinate will be returned.
#'
#' @param crs The coordinate reference system (CRS) of the data, passed to
#'   [sf::st_crs()]. By default this is [EPSG:4326](https://epsg.io/4326), the
#'   CRS associated with the commonly used latitude and longitude coordinates.
#'   Different coordinate systems can be specified using `crs` (e.g., `crs =
#'   27700` for the [British National Grid](https://epsg.io/27700)). Note that
#'   non-lat/lng coordinate systems will be re-projected to `EPSG:4326` for
#'   making comparisons with the NOAA metadata.
#'
#' @param database One of `"hourly"` or `"daily"`, which defines whether to
#'   import stations available in the GHCNh or GHCNd. Note that there is overlap
#'   between the two, but some stations may only be available in one or the
#'   other.
#'
#' @param provider When `return = "map"`, by default a map will be created in
#'   which readers may toggle between a vector street map and a satellite/aerial
#'   image. `provider` allows users to override this default; see
#'   \url{http://leaflet-extras.github.io/leaflet-providers/preview/} for a list
#'   of all base maps that can be used. Base maps can be toggled using a layer
#'   control menu; the labels will be taken from the name of the base map unless
#'   a named list is defined (see default value).
#'
#' @param return The type of R object to import the data as. One of the
#'   following:
#'
#' - `"table"`, which returns an R `data.frame`.
#'
#' - `"sf"`, which returns a spatial `data.frame` from the `sf` package.
#'
#' - `"map"`, which returns an interactive `leaflet` map.
#'
#' @author Jack Davison
#' @family GHCN functions
#'
#' @return One of:
#'
#' - a [tibble][tibble::tibble-package]
#'
#' - an [sf][sf::st_as_sf()] object
#'
#' - an interactive `leaflet` map
#'
#' @export
import_ghcn_stations <-
  function(
    name = NULL,
    country = NULL,
    state = NULL,
    lat = NULL,
    lng = NULL,
    crs = 4326,
    n_max = 10L,
    provider = c(
      "Street Map" = "CartoDB.Voyager",
      "Satellite" = "Esri.WorldImagery"
    ),
    database = c("hourly", "daily"),
    return = c("table", "sf", "map")
  ) {
    database <- rlang::arg_match(database, c("hourly", "daily"))
    return <- rlang::arg_match(return, c("table", "sf", "map"))
    meta_url <-
      switch(
        database,
        "hourly" = "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/doc/ghcnh-station-list.txt",
        "daily" = "https://www.ncei.noaa.gov/pub/data/ghcn/daily/ghcnd-stations.txt"
      )

    meta <-
      readr::read_fwf(
        meta_url,
        col_positions = readr::fwf_positions(
          start = c(1, 13, 22, 32, 39, 42, 73, 77, 81),
          end = c(11, 20, 30, 37, 40, 71, 75, 79, NA),
          col_names = c(
            "id",
            "lat",
            "lng",
            "elevation",
            "state",
            "name",
            "gsn_flag",
            "hcn_crn_flag",
            "wmo_id"
          )
        ),
        col_types = list(
          readr::col_character(),
          readr::col_number(),
          readr::col_number(),
          readr::col_number(),
          readr::col_character(),
          readr::col_character(),
          readr::col_character(),
          readr::col_character(),
          readr::col_character()
        ),
        na = c("-999.9", "-999", "-999.0", ""),
        progress = FALSE
      )

    meta <-
      dplyr::mutate(
        meta,
        country = trimws(substr(.data$id, 1L, 2L)),
        network = trimws(substr(.data$id, 3L, 3L)),
        .after = "id"
      ) |>
      dplyr::relocate(dplyr::all_of(c(
        "id",
        "name",
        "country",
        "state",
        "network"
      )))

    if (!is.null(name)) {
      meta <- dplyr::filter(meta, grepl(!!name, .data$name, ignore.case = TRUE))
    }

    if (!is.null(country)) {
      meta <- dplyr::filter(meta, .data$country %in% !!country)
    }

    if (!is.null(state)) {
      meta <- dplyr::filter(meta, .data$state %in% !!state)
    }

    if (!is.null(lat) && !is.null(lng)) {
      meta_sf <- sf::st_as_sf(
        meta,
        coords = c("lng", "lat"),
        crs = 4326,
        remove = FALSE
      )

      target_sf <-
        sf::st_as_sf(
          tibble::tibble(lat = lat, lng = lng),
          coords = c("lng", "lat"),
          crs = crs
        ) |>
        sf::st_transform(crs = 4326)

      meta_sf$distance <- as.numeric(sf::st_distance(meta_sf, target_sf)) /
        1000L

      meta <-
        meta_sf |>
        dplyr::arrange(.data$distance) |>
        dplyr::slice_head(n = n_max) |>
        tibble::tibble() |>
        dplyr::select(-"geometry")
    }

    if (return %in% c("sf", "map")) {
      meta <- sf::st_as_sf(
        meta,
        coords = c("lng", "lat"),
        crs = 4326,
        remove = FALSE
      )

      if (return == "map") {
        rlang::check_installed("leaflet")

        fmt_val <- function(x) {
          ifelse(
            is.na(x),
            "<span style='color: #bbb;'>N/A</span>",
            as.character(x)
          )
        }

        popup <- paste0(
          "<div style='font-family: Arial, sans-serif; min-width: 220px; max-width: 280px;'>",

          # Header
          "<div style='background: #2c7bb6; color: white; padding: 8px 12px; margin: -10px -10px 10px; border-radius: 4px 4px 0 0;'>",
          "<div style='font-size: 14px; font-weight: bold; margin-bottom: 4px;'>",
          meta$name,
          "</div>",
          "<div style='font-family: monospace; font-size: 12px; background: rgba(0,0,0,0.2); display: inline-block; padding: 2px 6px; border-radius: 3px; letter-spacing: 0.5px;'>",
          meta$id,
          "</div>",
          "</div>",

          # Geography section
          "<div style='margin-bottom: 8px;'>",
          "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Geography</div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Country</span><span>",
          fmt_val(meta$country),
          "</span></div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>State</span><span>",
          fmt_val(meta$state),
          "</span></div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Network</span><span>",
          fmt_val(meta$network),
          "</span></div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Elevation</span><span>",
          fmt_val(meta$elevation),
          " m</span></div>",
          "</div>",

          # Flags section
          "<div>",
          "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Station Flags</div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>GSN Flag</span><span>",
          fmt_val(meta$gsn_flag),
          "</span></div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>HCN/CRN Flag</span><span>",
          fmt_val(meta$hcn_crn_flag),
          "</span></div>",
          "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>WMO ID</span><span>",
          fmt_val(meta$wmo_id),
          "</span></div>",
          "</div>",

          "</div>"
        )
        overlays <- c("Stations")
        if ("distance" %in% names(meta)) {
          popup <- paste0(
            popup,
            "<div style='margin-top: 8px;'>",
            "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Distance from search</div>",
            "<div style='font-family: monospace; font-size: 12px; background: #f4f4f4; border-left: 3px solid #c0392b; padding: 3px 8px; border-radius: 0 3px 3px 0;'>",
            round(meta$distance, 1),
            " km",
            "</div>",
            "</div>",
            "</div>" # closes the outer wrapper div
          )
          overlays <- c(overlays, "Target")
        }

        n_sites <- nrow(meta)
        meta <- leaflet::leaflet(meta)

        if (!rlang::is_named(provider)) {
          provider <- stats::setNames(provider, provider)
        }
        for (i in seq_along(provider)) {
          meta <- leaflet::addProviderTiles(
            meta,
            provider = provider[[i]],
            group = names(provider)[[i]]
          )
        }

        meta <- meta |>
          leaflet::addMarkers(
            popup = popup,
            clusterOptions = if (n_sites < 20) {
              NULL
            } else {
              leaflet::markerClusterOptions()
            },
            group = "Stations"
          ) |>
          leaflet::addLayersControl(
            baseGroups = names(provider),
            overlayGroups = overlays,
            options = leaflet::layersControlOptions(
              collapsed = FALSE,
              autoZIndex = TRUE
            )
          )

        if ("Target" %in% overlays) {
          meta <-
            leaflet::addAwesomeMarkers(
              map = meta,
              data = target_sf,
              popup = paste0(
                "<div style='font-family: Arial, sans-serif; min-width: 200px; max-width: 260px;'>",

                # Header
                "<div style='background: #c0392b; color: white; padding: 8px 12px; margin: -10px -10px 10px; border-radius: 4px 4px 0 0;'>",
                "<div style='font-size: 14px; font-weight: bold;'>Search Location</div>",
                "</div>",

                # Coordinates section
                "<div style='margin-bottom: 8px;'>",
                "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Coordinates</div>",
                "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Latitude</span><span style='font-family: monospace;'>",
                lat,
                "</span></div>",
                "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Longitude</span><span style='font-family: monospace;'>",
                lng,
                "</span></div>",
                "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>CRS</span><span style='font-family: monospace;'>EPSG:",
                crs,
                "</span></div>",
                "</div>",

                "</div>"
              ),
              group = "Target",
              icon = leaflet::makeAwesomeIcon(
                icon = "circle",
                library = "fa",
                markerColor = "red",
                iconColor = "#FFFFFF"
              )
            )
        }
      }
    }

    return(meta)
  }

#' Import station inventory for the Global Historical Climatology Network
#'
#' This function accesses a data inventory of GHCN stations available through
#' either the GHCNh or GHCNd. The returned `data.frame` contains data which
#' reveals the earliest and latest years of data available for each station from
#' the NOAA database.
#'
#' @param database One of `"hourly"` or `"daily"`, which defines whether to
#'   import the GHCNh or GHCNd inventory. The way in which these files is
#'   formatted is different.
#'
#' @param pivot One of `"wide"` or `"long"`. The GHCNh inventory can be returned
#'   in a `"wide"` format (with `id`, `year` and twelve month columns) or a
#'   `"long"` format (with `id`, `year`, `month`, and `count` columns). Does not
#'   apply to the GHCNd inventory.
#'
#' @param progress The inventory file is large and can be slow to download. Show
#'   a progress indicator when accessing the inventory? Defaults to `TRUE` in
#'   interactive R sessions. Passed to `progress` in [readr::read_fwf()] and/or
#'   [purrr::pmap()].
#'
#' @author Jack Davison
#' @family GHCN functions
#'
#' @return a [tibble][tibble::tibble-package]
#' @export
import_ghcn_inventory <-
  function(
    database = c("hourly", "daily"),
    pivot = c("wide", "long"),
    progress = rlang::is_interactive()
  ) {
    database <- rlang::arg_match(database)
    pivot <- rlang::arg_match(pivot)

    if (database == "daily") {
      inventory <-
        readr::read_fwf(
          file = "https://www.ncei.noaa.gov/pub/data/ghcn/daily/ghcnd-inventory.txt",
          col_positions = readr::fwf_widths(
            c(11L, 9L, 10L, 5L, 5L, 5L),
            col_names = c(
              "id",
              "lat",
              "lng",
              "element",
              "start_year",
              "end_year"
            )
          ),
          show_col_types = FALSE,
          progress = progress
        ) |>
        dplyr::mutate(
          start_year = as.integer(.data$start_year),
          end_year = as.integer(.data$end_year)
        ) |>
        dplyr::mutate(
          country = trimws(substr(.data$id, 1L, 2L)),
          network = trimws(substr(.data$id, 3L, 3L)),
          .after = "id"
        )
    }

    if (database == "hourly") {
      temp_inv <- tempfile(fileext = "txt")
      utils::download.file(
        "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/doc/ghcnh-inventory.txt",
        destfile = temp_inv,
        quiet = !progress,
        mode = "wb"
      )

      inventory <- readr::read_fwf(
        temp_inv,
        show_col_types = FALSE,
        progress = progress
      )
      inventory <- dplyr::slice_tail(
        stats::setNames(inventory, as.vector(t(inventory[1, ]))),
        n = -1
      )
      inventory <-
        dplyr::mutate(
          inventory,
          dplyr::across("YEAR":"DEC", as.integer)
        )

      if (pivot == "wide") {
        inventory <-
          inventory |>
          dplyr::mutate(dplyr::across(-"GHCNh_ID", as.integer)) |>
          dplyr::rename_with(tolower) |>
          dplyr::rename("id" = "ghcnh_id")
      }

      if (pivot == "long") {
        inventory <-
          inventory |>
          tidyr::pivot_longer(
            "JAN":"DEC",
            names_to = "month",
            values_to = "count"
          ) |>
          dplyr::rename_with(tolower) |>
          dplyr::mutate(
            month = factor(
              .data$month,
              levels = c(
                "JAN",
                "FEB",
                "MAR",
                "APR",
                "MAY",
                "JUN",
                "JUL",
                "AUG",
                "SEP",
                "OCT",
                "NOV",
                "DEC"
              )
            ),
            dplyr::across(c("year", "count"), as.integer)
          ) |>
          dplyr::rename("id" = "ghcnh_id")
      }

      inventory <-
        dplyr::mutate(
          inventory,
          country = trimws(substr(.data$id, 1L, 2L)),
          network = trimws(substr(.data$id, 3L, 3L)),
          .after = "id"
        )
    }

    return(inventory)
  }

#' Import FIPS country codes and State/Province/Territory codes used by the
#' Global Historical Climatology Network
#'
#' This function returns a two-column dataframe either of "Federal Information
#' Processing Standards" (FIPS) codes and the countries to which they are
#' associated, or state codes and their associated states. These may be a useful
#' reference when examining GHCN site metadata.
#'
#' @param table One of `"countries"` or `"states"`.
#'
#' @param database One of `"hourly"`, `"daily"` or `"monthly"`, which defines
#'   which of the NOAA databases to import the FIPS codes from. There is little
#'   difference between the data in the different sources, but this option may
#'   be useful if one of the services is not accessible.
#'
#' @author Jack Davison
#' @family GHCN functions
#'
#' @return a [tibble][tibble::tibble-package]
#' @export
import_ghcn_countries <-
  function(
    table = c("countries", "states"),
    database = c("hourly", "daily", "monthly")
  ) {
    table <- rlang::arg_match(table)
    if (table == "countries") {
      database <- rlang::arg_match(database)
      url <- switch(
        database,
        "hourly" = "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/doc/ghcn-countries.txt",
        "daily" = "https://www.ncei.noaa.gov/pub/data/ghcn/daily/ghcnd-countries.txt",
        "monthly" = "https://www.ncei.noaa.gov/pub/data/ghcn/v4/ghcnm-countries.txt"
      )
    }

    if (table == "states") {
      url <-
        "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/doc/ghcn-states.txt"
    }

    x <- readr::read_lines(url, progress = FALSE)

    data <-
      tibble::tibble(
        code = trimws(substr(x, 1, 2)),
        x = trimws(substr(x, 3, 1000))
      ) |>
      dplyr::arrange(.data$code)

    newname <- ifelse(table == "countries", "country", "state")
    names(data)[names(data) == "x"] <- newname

    return(data)
  }

#' Import data from the Global Historical Climatology hourly (GHCNh) database
#'
#' This function flexibly accesses meteorological data from the GHCNh database.
#' Users can provide any number of years and stations, and fully control the
#' sorts of data flag codes that are returned with the data. By default, column
#' names are shortened for easier use in R, but longer, more descriptive names
#' can be requested.
#'
#' @section Data Definition:
#'
#'   The following core elements are in the GHCNh:
#'
#'   - **wind_direction:** (`wd`) Wind direction from true north (degrees). 360 = north, 180 = south, 270 = west; 000 indicates calm winds.
#'
#'   - **wind_speed:** (`ws`) Average wind speed (m/s).
#'
#'   - **temperature:** (`air_temp`) Air (dry bulb) temperature at approximately 2 meters above ground level, in degrees Celsius (to tenths).
#'
#'   - **station_level_pressure:** (`atmos_pres`) Pressure observed at station elevation; true barometric pressure at the location (hPa).
#'
#'   - **visibility:** (`visibility`) Horizontal visibility distance (km).
#'
#'   - **dew_point_temperature:** (`dew_point`) Dew point temperature in degrees Celsius (to tenths).
#'
#'   - **relative_humidity:** (`rh`) Relative humidity in whole percent, measured or calculated from temperature and dew point.
#'
#'   - **sky_cover:** (`cl`) Maximum of all **sky_cover_X** elements (oktas).
#'
#'   - **sky_cover_baseht:** (`cl_baseht`) The height above ground level (AGL) of the lowest cloud or obscuring phenomena layer aloft with 5/8 or more summation total sky cover, which may be predominantly opaque, or the vertical visibility into a surface-based obstruction.
#'
#'   - **sky_cover_X:** (`cl_X`) Fraction of sky covered by clouds (oktas). Up to 3 layers reported.
#'
#'   - **sky_cover_baseht_X:** (`cl_baseht_X`) Cloud base height for lowest layer (m). Up to 3 layers reported.
#'
#'   - **precipitation:** (`precip`) Total liquid precipitation (rain or melted snow) for the hour (mm). "T" indicates trace amounts.
#'
#'   - **precipitation_XX_hour:** (`precip_XX`) 3-hour total liquid precipitation (mm). "T" indicates trace amounts.
#'
#'   When `extra = TRUE`, the following additional columns are included:
#'
#'   - **sea_level_pressure:** (`sea_pres`) Estimated pressure at sea level directly below the station using actual temperature profile (hPa).
#'
#'   - **wind_gust:** (`wg`) Peak short-duration wind speed (m/s) exceeding the average wind speed.
#'
#'   - **wet_bulb_temperature:** (`wet_bulb`) Wet bulb temperature in degrees Celsius (to tenths), measured or calculated from temperature, dew point, and station pressure.
#'
#'   - **snow_depth:** (`snow_depth`) Depth of snowpack on the ground (mm).
#'
#'   - **altimeter:** (`altimeter`) Pressure reduced to mean sea level using standard atmosphere profile (hPa).
#'
#'   - **pressure_3hr_change:** (`pres_03`) Change in atmospheric pressure over a 3-hour period (hPa), with tendency code.
#'
#'   If `hourly = FALSE`, the following character columns may also be present.
#'
#'   - **pres_wx_MWX:** (`wx_mwX`) Present weather observation from manual reports (code). Up to 3 observations per report.
#'
#'   - **pres_wx_AUX:** (`wx_auX`) Present weather observation from automated ASOS/AWOS sensors (code). Up to 3 observations per report.
#'
#'   - **pres_wx_AWX:** (`wx_aqX`) Present weather observation from automated sensors (code). Up to 3 observations per report.
#'
#'   - **remarks:** (`remarks`) Raw observation remarks encoded in ICAO-standard METAR/SYNOP format.
#'
#' @param station One or more site codes for the station(s) of interest,
#'   obtained using [import_ghcn_stations()].
#'
#' @param year One or more years of interest. If `NULL`, the default, all years
#'   of data available for the chosen `station`s will be imported. Note that, in
#'   the GHCNd and GHCNm, files are split by station but not year, so setting a
#'   `year` will not speed up the download. Specifying fewer years will improve
#'   the speed of a GHCNh download, however.
#'
#' @param source There are two identical data formats to read from - `"psv"`
#'   (flat, pipe-delimited files) and `"parquet"` (a newer, faster, columnar
#'   format). The latter is typically faster, but requires the `arrow` package
#'   as an additional dependency. Note that this only applies when `year` is not
#'   `NULL`; all `by-site` files are `psv` files.
#'
#' @param hourly Should hourly means be calculated? The default is `TRUE`. If
#'   `FALSE` then the raw data are returned, which can be sub-hourly.
#'
#' @param extra Should additional columns be returned? The default, `FALSE`,
#'   returns an opinionated selection of elements that'll be of most interest to
#'   most users. `TRUE` will return everything available.
#'
#' @param abbr_names Should column names be abbreviated? When `TRUE`, the
#'   default, columns like `"wind_direction"` are shortened to `"wd"`. When
#'   `FALSE`, names will match the raw data, albeit in lower case.
#'
#' @param append_codes Logical. Should various codes and flags be appended to
#'   the output dataframe?
#'
#' @param codes When `append_codes` is `TRUE`, which codes should be appended to
#'   the dataframe? Any combination of `"measurement_code"`, `"quality_code"`,
#'   `"report_type"`, `"source_code"`, and/or `"source_id"`.
#'
#' @param progress Show a progress bar when importing many stations/years?
#'   Defaults to `TRUE` in interactive R sessions. Passed to `.progress` in
#'   [purrr::map()] and/or [purrr::pmap()].
#'
#' @section Parallel Processing:
#'
#'   If you are importing a lot of meteorological data, this can take a long
#'   while. This is because each combination of year and station requires
#'   downloading a separate data file from NOAA's online data directory, and the
#'   time each download takes can quickly add up. Many data import functions in
#'   `{worldmet}` can use parallel processing to speed downloading up, powered
#'   by the capable `{mirai}` package. If users have any `{mirai}` "daemons"
#'   set, these functions will download files in parallel. The greatest benefits
#'   will be seen if you spawn as many daemons as you have cores on your
#'   machine, although one fewer than the available cores is often a good rule
#'   of thumb. Your mileage may vary, however, and naturally spawning more
#'   daemons than station-year combinations will lead to diminishing returns.
#'
#'   ```
#'   # set workers - once per session
#'   mirai::daemons(4)
#'
#'   # import lots of data - NB: no change in the import function!
#'   big_met <- import_ghcn_hourly(code = "UKI0000EGLL", year = 2010:2025)
#'   ```
#'
#' @author Jack Davison
#' @family GHCN functions
#'
#' @return a [tibble][tibble::tibble-package]
#' @export
import_ghcn_hourly <-
  function(
    station = "UKI0000EGLL",
    year = NULL,
    source = c("psv", "parquet"),
    hourly = TRUE,
    extra = FALSE,
    abbr_names = TRUE,
    append_codes = FALSE,
    codes = c(
      "measurement_code",
      "quality_code",
      "report_type",
      "source_code",
      "source_id"
    ),
    progress = rlang::is_interactive()
  ) {
    # check source option
    source <- rlang::arg_match(source, c("psv", "parquet"), multiple = FALSE)
    if (source == "parquet") {
      rlang::check_installed("arrow")
    }

    # check codes
    potential_codes <- c(
      "measurement_code",
      "quality_code",
      "report_type",
      "source_code",
      "source_id"
    )
    codes <- rlang::arg_match(codes, potential_codes, multiple = TRUE)

    # (temporary?) issue with parquet files not having ids/lats/longs - need to
    # have meta to sort that out
    meta <- import_ghcn_stations(database = "hourly", return = "table")

    # iterate - need two strategies: one for year=NULL, one for station&year
    if (is.null(year)) {
      data <- purrr::map(
        .x = station,
        .f = purrr::in_parallel(
          .f = \(x) {
            import_single_ghcn_site(
              station = x,
              year = year,
              source = source,
              append_codes = append_codes,
              codes = codes,
              potential_codes = potential_codes,
              meta = meta
            )
          },
          import_single_ghcn_site = import_single_ghcn_site,
          year = year,
          source = source,
          append_codes = append_codes,
          codes = codes,
          potential_codes = potential_codes,
          meta = meta
        ),
        .progress = progress
      )
    } else {
      combos <- tidyr::crossing(station = station, year = year)
      data <- purrr::map2(
        .x = combos$station,
        .y = combos$year,
        .f = purrr::in_parallel(
          .f = \(x, y) {
            import_single_ghcn_site(
              station = x,
              year = y,
              source = source,
              append_codes = append_codes,
              codes = codes,
              potential_codes = potential_codes,
              meta = meta
            )
          },
          import_single_ghcn_site = import_single_ghcn_site,
          source = source,
          append_codes = append_codes,
          codes = codes,
          potential_codes = potential_codes,
          meta = meta
        ),
        .progress = progress
      )
    }

    # bind data together
    data <- dplyr::bind_rows(data)

    # if null, leave early
    if (is.null(data) || nrow(data) == 0L) {
      cli::cli_warn("No data has been returned.")
      return(NULL)
    }

    # scrub columns if not `extra`
    if (!extra) {
      data <-
        dplyr::select(
          data,
          "station_id",
          "station_name",
          "date",
          "latitude",
          "longitude",
          "elevation",
          dplyr::starts_with("wind_direction"),
          dplyr::starts_with("wind_speed"),
          dplyr::starts_with("temperature"),
          dplyr::starts_with("station_level_pressure"),
          dplyr::starts_with("visibility"),
          dplyr::starts_with("dew_point_temperature"),
          dplyr::starts_with("relative_humidity"),
          dplyr::starts_with("sky_cover"),
          dplyr::starts_with("precip")
        )
    }

    # function to rename columns
    rn_data <- function(data, from, to) {
      names(data) <- gsub(from, to, names(data))
      data
    }

    # if abbreviated names requested, do that
    if (abbr_names) {
      data <-
        data |>
        rn_data("latitude", "lat") |>
        rn_data("longitude", "lng") |>
        rn_data("elevation", "elev") |>
        rn_data("\\btemperature", "air_temp") |>
        rn_data("dew_point_temperature", "dew_point") |>
        rn_data("station_level_pressure", "atmos_pres") |>
        rn_data("sea_level_pressure", "sea_pres") |>
        rn_data("wind_direction", "wd") |>
        rn_data("wind_speed", "ws") |>
        rn_data("wind_gust", "wg") |>
        rn_data("relative_humidity", "rh") |>
        rn_data("wet_bulb_temperature", "wet_bulb") |>
        rn_data("wet_bulb_temperature", "wet_bulb") |>
        rn_data("pressure_3hr_change", "pres_03") |>
        dplyr::rename_with(\(x) gsub("precipitation", "precip", x)) |>
        dplyr::rename_with(\(x) gsub("pres_wx_", "wx_", x)) |>
        dplyr::rename_with(\(x) gsub("_hour", "", x)) |>
        dplyr::rename_with(\(x) gsub("sky_cover", "cl", x)) |>
        rn_data("precip_3", "precip_03") |>
        rn_data("precip_6", "precip_06") |>
        rn_data("precip_9", "precip_09") |>
        dplyr::rename_with(\(x) gsub("_measurement_code", "_mc", x)) |>
        dplyr::rename_with(\(x) gsub("_quality_code", "_qc", x)) |>
        dplyr::rename_with(\(x) gsub("_report_type", "_rt", x)) |>
        dplyr::rename_with(\(x) gsub("_source_code", "_sc", x)) |>
        dplyr::rename_with(\(x) gsub("_source_id", "_si", x)) |>
        dplyr::rename_with(\(x) gsub("_source_station_id", "_si", x))

      names(data)
    }

    # time average to hourly concentrations
    if (hourly) {
      # if names not abbreviated, switch to ws/wd (needed for timeAverage)
      if (!abbr_names) {
        data <-
          data |>
          rn_data("wind_direction", "wd") |>
          rn_data("wind_speed", "ws")
      }

      # time average to hourly
      data <- worldmet_time_average(
        data,
        avg.time = "hour",
        type = c("station_id", "station_name")
      )

      # if names not abbreviated, switch back to defaults
      if (!abbr_names) {
        data <-
          data |>
          rn_data("wd", "wind_direction") |>
          rn_data("ws", "wind_speed")
      }
    }

    return(data)
  }

#' Helper to import a single site
#' @noRd
import_single_ghcn_site <- function(
  station,
  year,
  source,
  append_codes,
  codes,
  potential_codes,
  meta
) {
  if (is.null(year)) {
    per_station_url <-
      "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/access/by-station/GHCNh_INPUTCODE_por.psv"

    url <- sub("INPUTCODE", station, per_station_url)

    data <- try(
      suppressWarnings(
        readr::read_delim(
          url,
          delim = "|",
          progress = FALSE,
          show_col_types = FALSE
        )
      ),
      silent = TRUE
    )

    if (class(data)[1] == "try-error") {
      return(NULL)
    }

    data <- data |>
      dplyr::slice_head(n = -1L) |>
      dplyr::rename_with(tolower) |>
      dplyr::mutate(
        date = ISOdate(
          year = .data$year,
          month = .data$month,
          day = .data$day,
          hour = .data$hour,
          min = .data$minute,
          sec = 0,
          tz = "UTC"
        ),
        .before = year,
        .keep = "unused"
      )
  } else {
    url <-
      paste0(
        "https://www.ncei.noaa.gov/oa/global-historical-climatology-network/hourly/access/by-year/INPUTYEAR/",
        source,
        "/GHCNh_INPUTCODE_INPUTYEAR.",
        source
      )

    url <- sub("INPUTCODE", station, url)
    url <- gsub("INPUTYEAR", as.character(year), url)

    if (source == "psv") {
      data <- try(
        suppressWarnings(
          readr::read_delim(
            url,
            delim = "|",
            name_repair = tolower,
            show_col_types = FALSE,
            progress = FALSE
          )
        ),
        silent = TRUE
      )

      if (class(data)[1] == "try-error") {
        return(NULL)
      }

      data <- dplyr::rename(data, station_id = "station")
    } else if (source == "parquet") {
      data <- try(
        suppressWarnings(
          arrow::read_parquet(url)
        ),
        silent = TRUE
      )

      if (class(data)[1] == "try-error") {
        return(NULL)
      }

      data <- data |>
        dplyr::rename_with(tolower) |>
        dplyr::mutate(
          date = as.POSIXct(
            .data$date,
            format = "%Y-%m-%dT%H:%M:%S",
            tz = "UTC"
          )
        )
    }
  }

  # missing wind directions
  if ("wind_direction" %in% names(data)) {
    data$wind_direction <- ifelse(
      data$wind_direction == 999,
      NA,
      data$wind_direction
    )
  }

  # coalesce sky covers
  # coalesce sky covers
  if ("sky_cover_layer_1" %in% names(data)) {
    data <- data |>
      tidyr::separate_wider_delim(
        cols = dplyr::matches("^sky_cover_layer_[123456789]$"),
        delim = ":",
        names_sep = "_",
        names = c("code", "okta"),
        too_few = "align_start"
      ) |>
      dplyr::select(
        -dplyr::matches("^sky_cover_layer_[123456789]_code$")
      ) |>
      dplyr::mutate(
        dplyr::across(
          dplyr::matches("^sky_cover_layer_[123456789]_okta$"),
          as.numeric
        )
      )

    # derive total sky cover from sky_condition column
    # format is up to 3 values separated by ";"; take the first non-empty one
    if ("sky_condition" %in% names(data)) {
      sky_parts <- strsplit(data$sky_condition, ";", fixed = TRUE)
      first_val <- vapply(
        sky_parts,
        function(p) {
          non_empty <- p[nzchar(p)]
          if (length(non_empty) == 0L) NA_character_ else non_empty[1L]
        },
        character(1L)
      )
      data$sky_cover <- as.integer(first_val)
      data$sky_cover[
        !is.na(data$sky_cover) & data$sky_cover == 9L
      ] <- NA_integer_
    }

    # get ceiling height
    data <-
      dplyr::mutate(
        data,
        sky_cover_baseht = dplyr::case_when(
          .data$sky_cover_layer_1_okta >= 5 ~ .data$sky_cover_layer_baseht_1,
          .data$sky_cover_layer_2_okta >= 5 ~ .data$sky_cover_layer_baseht_2,
          .data$sky_cover_layer_3_okta >= 5 ~ .data$sky_cover_layer_baseht_3,
          .default = NA
        )
      )

    # move columns
    data <-
      dplyr::relocate(
        data,
        "sky_cover",
        "sky_cover_baseht",
        .before = "sky_cover_layer_1_okta"
      )
  }

  # ensure data type consistency
  data <-
    dplyr::mutate(
      data,
      # all the "codes" are characters
      dplyr::across(
        dplyr::ends_with(
          potential_codes
        ),
        as.character
      ),
      # these variables are written observations/codes
      dplyr::across(
        dplyr::any_of(
          c(
            "pres_wx_mw1",
            "pres_wx_mw2",
            "pres_wx_mw3",
            "pres_wx_au1",
            "pres_wx_au2",
            "pres_wx_au3",
            "pres_wx_aw1",
            "pres_wx_aw2",
            "pres_wx_aw3",
            "remarks"
          )
        ),
        as.character
      ),
      # these variables are all numeric
      dplyr::across(
        dplyr::any_of(
          c(
            "latitude",
            "longitude",
            "elevation",
            "temperature",
            "dew_point_temperature",
            "station_level_pressure",
            "sea_level_pressure",
            "wind_direction",
            "wind_speed",
            "wind_gust",
            "precipitation",
            "relative_humidity",
            "wet_bulb_temperature",
            "snow_depth",
            "visibility",
            "altimeter",
            "pressure_3hr_change",
            "sky_cover_baseht_1",
            "sky_cover_baseht_2",
            "sky_cover_baseht_3",
            "precipitation_3_hour",
            "precipitation_6_hour",
            "precipitation_9_hour",
            "precipitation_12_hour",
            "precipitation_15_hour",
            "precipitation_18_hour",
            "precipitation_21_hour",
            "precipitation_24_hour"
          )
        ),
        as.numeric
      )
    )

  # if appending codes, find out which ones we are dropping, else drop
  # them all
  if (append_codes) {
    codes_to_drop <- setdiff(potential_codes, codes)
  } else {
    codes_to_drop <- potential_codes
  }

  if (length(codes_to_drop) > 1L) {
    # sometimes its source_id, sometimes its source_station_id
    if ("source_id" %in% codes_to_drop) {
      codes_to_drop <- c(codes_to_drop, "source_station_id")
    }
    # drop each type of code
    for (i in codes_to_drop) {
      data <- dplyr::select(data, -dplyr::ends_with(i))
    }
  }

  # temporary issue w/ parquet files - no station_id lat/lng
  data$station_id <- station
  filtered_meta <- dplyr::filter(meta, .data$id == station)
  if (nrow(filtered_meta) == 1) {
    data$latitude <- filtered_meta$lat
    data$longitude <- filtered_meta$lng
  }

  # return data
  return(data)
}

#' Copy of [openair::timeAverage()] for hourly averaging, simplified for
#' specific use in worldmet
#' @noRd
worldmet_time_average <- function(
  mydata,
  avg.time = "hour",
  type = "default"
) {
  # extract variables of interest
  vars <- unique(c("date", names(mydata)))
  mydata <- check_prep(
    mydata,
    vars,
    type = "default",
    remove.calm = FALSE
  )

  # time zone of data (replace missing w/ GMT)
  TZ <- attr(mydata$date, "tzone") %||% "GMT"

  # function for getting mean
  FUN <- function(x) {
    if (all(is.na(x))) {
      NA
    } else {
      mean(x, na.rm = TRUE)
    }
  }

  # function to calculate means
  #
  # need to check whether avg.time is > or < actual time gap of data
  # then data will be expanded or aggregated accordingly
  calc.mean <- function(mydata) {
    # calculate Uu & Vv if "wd" (& "ws") are in mydata
    mydata <- calculate_wind_components(mydata = mydata)

    ## Aggregate data

    ## variables to split by
    vars <- c(type, "date")
    mydata$date <-
      cut(mydata$date, avg.time) |>
      as.character() |>
      (function(x) {
        id <- !grepl("00:00", x)
        x[id] <- paste(x[id], "00:00:00")
        x
      })() |>
      as.POSIXct(tz = TZ)

    avmet <-
      mydata |>
      dplyr::group_by(dplyr::across(dplyr::all_of(vars))) |>
      dplyr::summarise(dplyr::across(
        dplyr::where(function(x) {
          !is.factor(x) && !is.character(x)
        }),
        function(x) {
          FUN(x)
        }
      ))

    avmet
  }

  # type will always be a factor
  for (i in type) {
    mydata[i] <- factor(mydata[[i]])
  }

  # select date, type, and all non-factor/character columns
  mydata <-
    dplyr::select(
      mydata,
      dplyr::all_of(c("date", type)),
      dplyr::where(function(x) {
        !is.character(x) && !is.factor(x)
      })
    )

  # calculate averages
  mydata <-
    mydata |>
    split(mydata[type], drop = TRUE) |>
    purrr::map(
      calc.mean
    ) |>
    purrr::list_rbind() |>
    dplyr::as_tibble()

  if ("Uu" %in% names(mydata) && "Vv" %in% names(mydata)) {
    mydata <-
      mydata |>
      dplyr::mutate(
        wd = as.vector(atan2(.data$Uu, .data$Vv) * 360 / 2 / pi),
        wd = ifelse(.data$wd < 0, .data$wd + 360, .data$wd)
      ) |>
      dplyr::select(-dplyr::any_of(c("Uu", "Vv")))
  }

  # drop default column if it exists
  if ("default" %in% names(mydata)) {
    mydata$default <- NULL
  }

  # ensure all hours are present
  if (avg.time == "hour") {
    type_split <- c(type, "lat", "lng", "latitude", "longitude", "elev")
    type_split <- type_split[type_split %in% names(mydata)]

    # split by type, lat, lng, elev, and complete by date
    mydata <-
      purrr::map(
        .x = split(
          mydata,
          mydata[, type_split],
          drop = TRUE
        ),
        .f = \(df) {
          from <- as.POSIXct(
            paste0(format(min(df$date), "%Y"), "-01-01"),
            tz = attr(df$date, "tzone")
          )
          to <- as.POSIXct(
            paste0(format(max(df$date), "%Y"), "-12-31 23:00:00"),
            tz = attr(df$date, "tzone")
          )

          tidyr::complete(
            df,
            tidyr::nesting(!!!rlang::syms(type_split)),
            date = seq(from, to, by = "hour")
          )
        }
      ) |>
      dplyr::bind_rows()
  }

  # return
  return(mydata)
}

#' Calculate Uu and Vv if wd & ws are in the data
#' @noRd
calculate_wind_components <- function(mydata) {
  if ("wd" %in% names(mydata)) {
    if (is.numeric(mydata$wd)) {
      if ("ws" %in% names(mydata)) {
        mydata <- dplyr::mutate(
          mydata,
          Uu = .data$ws * sin(2 * pi * .data$wd / 360),
          Vv = .data$ws * cos(2 * pi * .data$wd / 360)
        )
      } else {
        mydata <- dplyr::mutate(
          mydata,
          Uu = sin(2 * pi * .data$wd / 360),
          Vv = cos(2 * pi * .data$wd / 360)
        )
      }
    }
  }

  return(mydata)
}

#' Copy of `check_prep()` from `openair`
#' @noRd
check_prep <- function(
  mydata,
  Names,
  type,
  remove.calm = TRUE,
  remove.neg = TRUE,
  wd = "wd"
) {
  # deal with conditioning variable if present, if user-defined, must exist in
  # data pre-defined types existing conditioning variables that only depend on
  # date (which is checked)
  conds <- c(
    "default",
    "year",
    "hour",
    "month",
    "season",
    "weekday",
    "week",
    "weekend",
    "monthyear",
    "gmtbst",
    "bstgmt",
    "dst",
    "daylight",
    "yearseason",
    "seasonyear"
  )
  all.vars <- unique(c(names(mydata), conds))

  # names we want to be there
  varNames <- c(Names, type)
  matching <- varNames %in% all.vars

  if (!all(matching)) {
    # not all variables are present
    stop(cat("Can't find the variable(s)", varNames[!matching], "\n"))
  }

  # add type to names if not in pre-defined list
  if (!all(type %in% conds)) {
    ids <- which(!type %in% conds)
    Names <- c(Names, type[ids])
  }

  # if type already present in data frame
  if (any(type %in% names(mydata))) {
    ids <- which(type %in% names(mydata))
    Names <- unique(c(Names, type[ids]))
  }

  # just select data needed
  mydata <- mydata[, Names]

  # if site is in the data set, check none are missing
  # seems to be a problem for some KCL data...
  if ("site" %in% names(mydata)) {
    # split by site

    # remove any NA sites
    if (anyNA(mydata$site)) {
      id <- which(is.na(mydata$site))
      mydata <- mydata[-id, ]
    }
  }

  # sometimes ratios are considered which can results in infinite values
  # make sure all infinite values are set to NA
  mydata[] <- lapply(mydata, function(x) {
    replace(x, x == Inf | x == -Inf, NA)
  })

  if ("ws" %in% Names) {
    if ("ws" %in% Names && is.numeric(mydata$ws)) {
      # check for negative wind speeds
      if (any(sign(mydata$ws[!is.na(mydata$ws)]) == -1)) {
        if (remove.neg) {
          # remove negative ws only if TRUE
          warning("Wind speed <0; removing negative data")
          mydata$ws[mydata$ws < 0] <- NA
        }
      }
    }
  }

  # round wd to make processing obvious
  # data already rounded to nearest 10 degress will not be affected
  # data not rounded will be rounded to nearest 10 degrees
  # assumes 10 is average of 5-15 etc
  if (wd %in% Names) {
    if (wd %in% Names && is.numeric(mydata[, wd])) {
      # check for wd <0 or > 360
      if (
        any(
          sign(mydata[[wd]][!is.na(mydata[[wd]])]) == -1 |
            mydata[[wd]][!is.na(mydata[[wd]])] > 360
        )
      ) {
        warning("Wind direction < 0 or > 360; removing these data")
        mydata[[wd]][mydata[[wd]] < 0] <- NA
        mydata[[wd]][mydata[[wd]] > 360] <- NA
      }

      if (remove.calm) {
        if ("ws" %in% names(mydata)) {
          mydata[[wd]][mydata$ws == 0] <- NA # set wd to NA where there are calms
          mydata$ws[mydata$ws == 0] <- NA # remove calm ws
        }
        mydata[[wd]][mydata[[wd]] == 0] <- 360 # set any legitimate wd to 360

        # round wd for use in functions - except windRose/pollutionRose
        mydata[[wd]] <- 10 * ceiling(mydata[[wd]] / 10 - 0.5)
        mydata[[wd]][mydata[[wd]] == 0] <- 360 # angles <5 should be in 360 bin
      }
      mydata[[wd]][mydata[[wd]] == 0] <- 360 # set any legitimate wd to 360
    }
  }

  # make sure date is ordered in time if present
  if ("date" %in% Names) {
    if (inherits(mydata$date, "POSIXlt")) {
      stop("date should be in POSIXct format not POSIXlt")
    }

    # try and work with a factor date - but probably a problem in original data
    if (is.factor(mydata$date)) {
      warning("date field is a factor, check date format")
      mydata$date <- as.POSIXct(mydata$date, "GMT")
    }

    mydata <- dplyr::arrange(mydata, date)

    # make sure date is the first field
    if (names(mydata)[1] != "date") {
      mydata <- mydata[c("date", setdiff(names(mydata), "date"))]
    }

    # check to see if there are any missing dates, stop if there are
    ids <- which(is.na(mydata$date))
    if (length(ids) > 0) {
      mydata <- mydata[-ids, ]
      warning(
        paste("Missing dates detected, removing", length(ids), "lines"),
        call. = FALSE
      )
    }
  }

  # return data frame
  return(mydata)
}

#' Import Meteorological data from the NOAA Integrated Surface Database (ISD)
#'
#' This is the main function to import data from the NOAA Integrated Surface
#' Database (ISD). The ISD contains detailed surface meteorological data from
#' around the world for over 30,000 locations.
#'
#' Note the following units for the main variables:
#'
#' \describe{
#'
#' \item{date}{Date/time in POSIXct format. **Note the time zone is GMT (UTC)
#' and may need to be adjusted to merge with other local data. See details
#' below.**}
#'
#' \item{latitude}{Latitude in decimal degrees (-90 to 90).}
#'
#' \item{longitude}{Longitude in decimal degrees (-180 to 180). Negative numbers
#' are west of the Greenwich Meridian.}
#'
#' \item{elevation}{Elevation of site in metres.}
#'
#' \item{wd}{Wind direction in degrees. 90 is from the east.}
#'
#' \item{ws}{Wind speed in m/s.}
#'
#' \item{ceil_hgt}{The height above ground level (AGL) of the lowest cloud or
#' obscuring phenomena layer aloft with 5/8 or more summation total sky cover,
#' which may be predominantly opaque, or the vertical visibility into a
#' surface-based obstruction.}
#'
#' \item{visibility}{The visibility in metres.}
#'
#' \item{air_temp}{Air temperature in degrees Celcius.}
#'
#' \item{dew_point}{The dew point temperature in degrees Celcius.}
#'
#' \item{atmos_pres}{The sea level pressure in millibars.}
#'
#' \item{RH}{The relative humidity (%).}
#'
#' \item{cl_1,  ...,  cl_3}{Cloud cover for different layers in Oktas (1-8).}
#'
#' \item{cl}{Maximum of cl_1 to cl_3 cloud cover in Oktas (1-8).}
#'
#' \item{cl_1_height, ..., cl_3_height}{Height of the cloud base for each later
#' in metres.}
#'
#' \item{precip_12}{12-hour precipitation in mm. The sum of this column should
#' give the annual precipitation.}
#'
#' \item{precip_6}{6-hour precipitation in mm.}
#'
#' \item{precip}{This value of precipitation spreads the 12-hour total across
#' the previous 12 hours.}
#'
#' \item{pwc}{The description of the present weather description (if
#' available).}
#'
#' }
#'
#' The data are returned in GMT (UTC). It may be necessary to adjust the time
#' zone when combining with other data. For example, if air quality data were
#' available for Beijing with time zone set to "Etc/GMT-8" (note the negative
#' offset even though Beijing is ahead of GMT. See the `openair` package and
#' manual for more details), then the time zone of the met data can be changed
#' to be the same. One way of doing this would be `attr(met$date, "tzone") <-
#' "Etc/GMT-8"` for a meteorological data frame called `met`. The two data sets
#' could then be merged based on `date`.
#'
#' @inheritSection import_ghcn_hourly Parallel Processing
#'
#' @param code The identifying code as a character string. The code is a
#'   combination of the USAF and the WBAN unique identifiers. The codes are
#'   separated by a \dQuote{-} e.g. `code = "037720-99999"`.
#'
#' @param year The year to import. This can be a vector of years e.g. `year =
#'   2000:2005`.
#'
#' @param hourly Should hourly means be calculated? The default is `TRUE`. If
#'   `FALSE` then the raw data are returned.
#'
#' @param source The NOAA ISD service stores files in two formats; as delimited
#'   CSV files (`"delim"`) and as fixed width files (`"fwf"`).
#'   [import_isd_hourly()] defaults to `"delim"` but, if the delimited data
#'   store is down, users may wish to try `"fwf"` instead. Both data sources
#'   should be identical to one another.
#'
#' @param progress Show a progress bar when importing many stations/years?
#'   Defaults to `TRUE` in interactive R sessions. Passed to `.progress` in
#'   [purrr::map()] and/or [purrr::pmap()].
#'
#' @param quiet Print missing sites/years to the screen? Defaults to `FALSE`.
#'
#' @export
#' @return Returns a data frame of surface observations. The data frame is
#'   consistent for use with the `openair` package. Note that the data are
#'   returned in GMT (UTC) time zone format. Users may wish to express the data
#'   in other time zones, e.g., to merge with air pollution data.
#'
#' @family NOAA ISD functions
#' @author David Carslaw
#' @examples
#'
#' \dontrun{
#' # import some data
#' beijing_met <- import_isd_hourly(code = "545110-99999", year = 2010:2011)
#'
#' # importing lots of data? use mirai for parallel processing
#' mirai::daemons(4)
#' beijing_met2 <- import_isd_hourly(code = "545110-99999", year = 2010:2025)
#' }
import_isd_hourly <- function(
  code = "037720-99999",
  year = 2014,
  hourly = TRUE,
  source = c("delim", "fwf"),
  progress = rlang::is_interactive(),
  quiet = FALSE
) {
  warn_isd()
  source <- rlang::arg_match(source)

  # main web site https://www.ncei.noaa.gov/products/land-based-station/integrated-surface-database
  # formats document https://www.ncei.noaa.gov/data/global-hourly/doc/isd-format-document.pdf
  # brief csv file description https://www.ncei.noaa.gov/data/global-hourly/doc/CSV_HELP.pdf
  # gis map https://gis.ncdc.noaa.gov/map/viewer/#app=cdo&cfg=cdo&theme=hourly&layers=1
  # go through each of the years selected, use parallel processing

  # sites and years to process
  site_process <- expand.grid(
    code = code,
    year = year,
    stringsAsFactors = FALSE
  )

  # read in files - in_parallel defaults to sequential if no cores, will
  # use parallelism if daemons are set
  if (source == "delim") {
    dat <-
      purrr::pmap(
        site_process,
        purrr::in_parallel(
          \(code, year) getDatDelim(code = code, year = year, hourly = hourly),
          getDatDelim = getDatDelim,
          hourly = hourly
        ),
        .progress = progress
      ) |>
      purrr::list_rbind()
  }

  if (source == "fwf") {
    dat <-
      purrr::pmap(
        site_process,
        purrr::in_parallel(
          \(code, year) {
            getDatFwf(
              code = code,
              year = year,
              hourly = hourly,
              precip = TRUE,
              PWC = TRUE
            )
          },
          getDatFwf = getDatFwf,
          hourly = hourly
        ),
        .progress = progress
      ) |>
      purrr::list_rbind()
  }

  if (is.null(dat) || nrow(dat) == 0) {
    # get source not used
    trysource <- c("fwf", "delim")
    trysource <- trysource[trysource != source]

    # message
    cli::cli_inform(
      c(
        "x" = "Specified {.field site}-{.field year} combinations do not exist.",
        "i" = "Is the ISD service down? Check {.url https://www.ncei.noaa.gov/data/global-hourly/}.",
        "i" = 'Try {.code import_isd_hourly(..., source = "{trysource}")} to access an alternative data store.'
      )
    )
    return()
  }

  # check to see what is missing and print to screen
  actual <- dplyr::select(dat, "code", "date", "station") |>
    dplyr::mutate(year = as.numeric(format(.data$date, "%Y"))) |>
    dplyr::slice(1, .by = c("code", "year"))

  actual <- dplyr::left_join(site_process, actual, by = c("code", "year"))

  if (length(which(is.na(actual$date))) > 0 && !quiet) {
    cli::cli_h1("The following sites / years are missing:")
    cli::cli_ul(id = "worldmetmissing")
    missing <- dplyr::filter(actual, is.na(.data$date))
    for (i in seq_len(nrow(missing))) {
      df_i <- missing[i, ]
      cli::cli_li(
        items = "{.field site:} {df_i$code} in {.field year:} {df_i$year}"
      )
    }
    cli::cli_end(id = "worldmetmissing")
  }

  return(dat)
}

#' @noRd
getDatDelim <- function(code, year, hourly) {
  ## location of data
  file.name <- paste0(
    "https://www.ncei.noaa.gov/data/global-hourly/access/",
    year,
    "/",
    gsub(pattern = "-", "", code),
    ".csv"
  )

  # suppress warnings because some fields might be missing in the list
  # Note that not all available data is returned - just what I think is most useful
  met_data <- try(
    suppressWarnings(readr::read_csv(
      file.name,
      col_types = readr::cols_only(
        STATION = readr::col_character(),
        DATE = readr::col_datetime(format = ""),
        SOURCE = readr::col_double(),
        LATITUDE = readr::col_double(),
        LONGITUDE = readr::col_double(),
        ELEVATION = readr::col_double(),
        NAME = readr::col_character(),
        REPORT_TYPE = readr::col_character(),
        CALL_SIGN = readr::col_double(),
        QUALITY_CONTROL = readr::col_character(),
        WND = readr::col_character(),
        CIG = readr::col_character(),
        VIS = readr::col_character(),
        TMP = readr::col_character(),
        DEW = readr::col_character(),
        SLP = readr::col_character(),
        AA1 = readr::col_character(),
        AW1 = readr::col_character(),
        GA1 = readr::col_character(),
        GA2 = readr::col_character(),
        GA3 = readr::col_character()
      ),
      progress = FALSE
    )),
    silent = TRUE
  )

  if (class(met_data)[1] == "try-error") {
    met_data <- NULL
    return()
  }

  met_data <- dplyr::rename(
    met_data,
    code = "STATION",
    station = "NAME",
    date = "DATE",
    latitude = "LATITUDE",
    longitude = "LONGITUDE",
    elev = "ELEVATION"
  )

  met_data$code <- code

  # separate WND column

  if ("WND" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "WND",
      into = c("wd", "x", "y", "ws", "z")
    )

    met_data <- dplyr::mutate(
      met_data,
      wd = as.numeric(.data$wd),
      wd = ifelse(.data$wd == 999, NA, .data$wd),
      ws = as.numeric(.data$ws),
      ws = ifelse(.data$ws == 9999, NA, .data$ws),
      ws = .data$ws / 10
    )
  }

  # separate TMP column
  if ("TMP" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "TMP",
      into = c("air_temp", "flag_temp"),
      sep = ","
    )

    met_data <- dplyr::mutate(
      met_data,
      air_temp = as.numeric(.data$air_temp),
      air_temp = ifelse(.data$air_temp == 9999, NA, .data$air_temp),
      air_temp = .data$air_temp / 10
    )
  }

  # separate VIS column
  if ("VIS" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "VIS",
      into = c("visibility", "flag_vis1", "flag_vis2", "flag_vis3"),
      sep = ",",
      fill = "right"
    )

    met_data <- dplyr::mutate(
      met_data,
      visibility = as.numeric(.data$visibility),
      visibility = ifelse(
        .data$visibility %in% c(9999, 999999),
        NA,
        .data$visibility
      )
    )
  }

  # separate DEW column
  if ("DEW" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "DEW",
      into = c("dew_point", "flag_dew"),
      sep = ","
    )

    met_data <- dplyr::mutate(
      met_data,
      dew_point = as.numeric(.data$dew_point),
      dew_point = ifelse(.data$dew_point == 9999, NA, .data$dew_point),
      dew_point = .data$dew_point / 10
    )
  }
  # separate SLP column
  if ("SLP" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "SLP",
      into = c("atmos_pres", "flag_pres"),
      sep = ",",
      fill = "right"
    )

    met_data <- dplyr::mutate(
      met_data,
      atmos_pres = as.numeric(.data$atmos_pres),
      atmos_pres = ifelse(
        .data$atmos_pres %in% c(99999, 999999),
        NA,
        .data$atmos_pres
      ),
      atmos_pres = .data$atmos_pres / 10
    )
  }

  # separate CIG (sky condition) column
  if ("CIG" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "CIG",
      into = c("ceil_hgt", "flag_sky1", "flag_sky2", "flag_sky3"),
      sep = ",",
      fill = "right"
    )

    met_data <- dplyr::mutate(
      met_data,
      ceil_hgt = as.numeric(.data$ceil_hgt),
      ceil_hgt = ifelse(.data$ceil_hgt == 99999, NA, .data$ceil_hgt)
    )
  }

  ## relative humidity - general formula based on T and dew point
  met_data$RH <- 100 *
    ((112 - 0.1 * met_data$air_temp + met_data$dew_point) /
      (112 + 0.9 * met_data$air_temp))^8

  if ("GA1" %in% names(met_data)) {
    # separate GA1 (cloud layer 1 height, amount) column
    met_data <- tidyr::separate(
      data = met_data,
      col = "GA1",
      into = c(
        "cl_1",
        "code_1",
        "cl_1_height",
        "code_2",
        "cl_1_type",
        "code_3"
      ),
      sep = ","
    )

    met_data <- dplyr::mutate(
      met_data,
      cl_1 = as.numeric(.data$cl_1),
      cl_1 = ifelse(
        (is.na(.data$cl_1) & .data$ceil_hgt == 22000),
        0,
        .data$cl_1
      ),
      cl_1 = ifelse(.data$cl_1 == 99, NA, .data$cl_1),
      cl_1_height = as.numeric(.data$cl_1_height),
      cl_1_height = ifelse(.data$cl_1_height == 99999, NA, .data$cl_1_height)
    )
  }

  if ("GA2" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "GA2",
      into = c(
        "cl_2",
        "code_1",
        "cl_2_height",
        "code_2",
        "cl_2_type",
        "code_3"
      ),
      sep = ","
    )

    met_data <- dplyr::mutate(
      met_data,
      cl_2 = as.numeric(.data$cl_2),
      cl_2 = ifelse(.data$cl_2 == 99, NA, .data$cl_2),
      cl_2_height = as.numeric(.data$cl_2_height),
      cl_2_height = ifelse(.data$cl_2_height == 99999, NA, .data$cl_2_height)
    )
  }

  if ("GA3" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "GA3",
      into = c(
        "cl_3",
        "code_1",
        "cl_3_height",
        "code_2",
        "cl_3_type",
        "code_3"
      ),
      sep = ","
    )

    met_data <- dplyr::mutate(
      met_data,
      cl_3 = as.numeric(.data$cl_3),
      cl_3 = ifelse(.data$cl_3 == 99, NA, .data$cl_3),
      cl_3_height = as.numeric(.data$cl_3_height),
      cl_3_height = ifelse(.data$cl_3_height == 99999, NA, .data$cl_3_height)
    )
  }

  ## for cloud cover, make new 'cl' max of 3 cloud layers
  if ("cl_3" %in% names(met_data)) {
    met_data$cl <- pmax(
      met_data$cl_1,
      met_data$cl_2,
      met_data$cl_3,
      na.rm = TRUE
    )
  }

  # PRECIP AA1
  if ("AA1" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "AA1",
      into = c("precip_code", "precip_raw", "code_1", "code_2"),
      sep = ","
    )

    met_data <- dplyr::mutate(
      met_data,
      precip_raw = as.numeric(.data$precip_raw),
      precip_raw = ifelse(.data$precip_raw == 9999, NA, .data$precip_raw),
      precip_raw = .data$precip_raw / 10
    )

    # deal with 6 and 12 hour precip
    id <- which(met_data$precip_code == "06")

    if (length(id) > 0) {
      met_data$precip_6 <- NA
      met_data$precip_6[id] <- met_data$precip_raw[id]
    }

    id <- which(met_data$precip_code == "12")

    if (length(id) > 0) {
      met_data$precip_12 <- NA
      met_data$precip_12[id] <- met_data$precip_raw[id]
    }
  }

  # weather codes, AW1

  if ("AW1" %in% names(met_data)) {
    met_data <- tidyr::separate(
      data = met_data,
      col = "AW1",
      into = c("pwc", "code_1"),
      sep = ",",
      fill = "right"
    )

    met_data <- dplyr::left_join(met_data, worldmet::weatherCodes, by = "pwc")
    met_data <- dplyr::select(met_data, -"pwc") |>
      dplyr::rename(pwc = "description")
  }

  ## select the variables we want
  met_data <- dplyr::select(
    met_data,
    dplyr::any_of(c(
      "date",
      "code",
      "station",
      "latitude",
      "longitude",
      "elev",
      "ws",
      "wd",
      "air_temp",
      "atmos_pres",
      "visibility",
      "dew_point",
      "RH",
      "ceil_hgt",
      "cl_1",
      "cl_2",
      "cl_3",
      "cl",
      "cl_1_height",
      "cl_2_height",
      "cl_3_height",
      "pwc",
      "precip_12",
      "precip_6",
      "precip"
    ))
  )

  ## present weather is character and cannot be averaged, take first
  if ("pwc" %in% names(met_data) && hourly) {
    pwc <- met_data[c("date", "pwc")]
    pwc$date2 <- format(pwc$date, "%Y-%m-%d %H") ## nearest hour
    tmp <- pwc[which(!duplicated(pwc$date2)), ]
    dates <- as.POSIXct(paste0(unique(pwc$date2), ":00:00"), tz = "GMT")

    pwc <- data.frame(date = dates, pwc = tmp$pwc)
  }

  ## average to hourly
  if (hourly) {
    met_data <-
      worldmet_time_average(
        met_data,
        avg.time = "hour",
        type = c("code", "station")
      )
  }

  ## add pwc back in
  if (exists("pwc")) {
    met_data <- dplyr::left_join(met_data, pwc, by = "date")
  }

  ## add precipitation - based on 12 HOUR averages, so work with hourly data

  ## spread out precipitation across each hour

  ## only do this if precipitation exists
  if ("precip_12" %in% names(met_data) && hourly) {
    ## make new precip variable
    met_data$precip <- NA

    ## id where there is 12 hour data
    id <- which(!is.na(met_data$precip_12))

    if (length(id) == 0L) {
      return()
    }

    id <- id[id > 11] ## make sure we don't run off beginning

    for (i in seq_along(id)) {
      met_data$precip[(id[i] - 11):id[i]] <- met_data$precip_12[id[i]] / 12
    }
  }

  # replace NaN with NA
  met_data[] <- lapply(met_data, function(x) {
    replace(x, is.nan(x), NA)
  })

  return(dplyr::as_tibble(met_data))
}

#' 'Fixed width file' helpers.
#' @noRd
getDatFwf <- function(code, year, hourly, precip, PWC) {
  procAdditFwf <- function(add, dat, precip, PWC) {
    # function to process additional data such as cloud cover

    # consider first 3 layers of cloud GA1, GA2, GA3
    dat <- extractCloudFwf(add, dat, "GA1", "cl_1")
    dat <- extractCloudFwf(add, dat, "GA2", "cl_2")
    dat <- extractCloudFwf(add, dat, "GA3", "cl_3")

    # 6 and 12 hour precipitation
    if (precip) {
      dat <- extractPrecipFwf(add, dat, "AA112", "precip_12")
      dat <- extractPrecipFwf(add, dat, "AA106", "precip_6")
    }

    if (PWC) {
      dat <- extractCurrentWeatherFwf(add, dat, "AW1")
    }

    return(dat)
  }

  extractPrecipFwf <- function(add, dat, field = "AA112", out = "precip_12") {
    # fields that contain search string
    id <- grep(field, add)

    # variables for precip amount
    dat[[out]] <- NA

    if (length(id) > 1) {
      # location of begining of AA1 etc

      loc <- sapply(id, function(x) regexpr(field, add[x]))

      # extract amount of rain

      amnt <- sapply(seq_along(id), function(x) {
        substr(add[id[x]], start = loc[x] + 5, stop = loc[x] + 8)
      })

      miss <- which(amnt == "9999") # missing
      if (length(miss) > 0) {
        amnt[miss] <- NA
      }

      amnt <- as.numeric(amnt) / 10

      dat[[out]][id] <- amnt
    }

    return(dat)
  }

  extractCloudFwf <- function(add, dat, field = "GA1", out = "cl_1") {
    # 3 fields are used: GA1, GA2 and GA3

    height <- paste0(out, "_height") # cloud height field

    # fields that contain search string
    id <- grep(field, add)

    # variables for cloud amount (oktas) and cloud height
    dat[[out]] <- NA
    dat[[height]] <- NA

    if (length(id) > 1) {
      # location of begining of GA1 etc

      loc <- sapply(id, function(x) regexpr(field, add[x]))

      # extract the variable
      cl <- sapply(seq_along(id), function(x) {
        substr(add[id[x]], start = loc[x] + 3, stop = loc[x] + 4)
      })
      cl <- as.numeric(cl)

      miss <- which(cl == 99) # missing

      if (length(miss) > 0) {
        cl[miss] <- NA
      }

      # and height of cloud
      h <- sapply(seq_along(id), function(x) {
        substr(add[id[x]], start = loc[x] + 6, stop = loc[x] + 11)
      })
      h <- as.numeric(h)

      miss <- which(h == 99999)
      if (length(miss) > 0) {
        h[miss] <- NA
      }

      dat[[out]][id] <- cl
      dat[[height]][id] <- h
    }

    return(dat)
  }

  extractCurrentWeatherFwf <- function(add, dat, field = "AW1") {
    # fields that contain search string
    id <- grep(field, add)

    if (length(id) > 1) {
      # name of output variable
      dat[["pwc"]] <- NA

      # location of begining of AW1
      loc <- sapply(id, function(x) regexpr(field, add[x]))

      # extract the variable
      pwc <- sapply(seq_along(id), function(x) {
        substr(add[id[x]], start = loc[x] + 3, stop = loc[x] + 4)
      })
      pwc <- as.character(pwc)

      # look up code in weatherCodes.RData

      desc <- sapply(pwc, function(x) {
        weatherCodes$description[which(weatherCodes$pwc == x)]
      })

      dat[["pwc"]][id] <- desc
    } else {
      return(dat)
    }

    return(dat)
  }

  weatherCodes <- worldmet::weatherCodes

  # location of data
  file.name <- paste0(
    "https://www.ncei.noaa.gov/pub/data/noaa/",
    year,
    "/",
    code,
    "-",
    year,
    ".gz"
  )

  # Download file to temp directory
  tmp <- paste0(tempdir(), basename(file.name))

  # deal with any missing data, issue warning
  bin <- try(utils::download.file(file.name, tmp, quiet = TRUE, mode = "wb"))

  if (inherits(bin, "try-error")) {
    return(NULL)
  }

  column_widths <- c(
    4,
    6,
    5,
    4,
    2,
    2,
    2,
    2,
    1,
    6,
    7,
    5,
    5,
    5,
    4,
    3,
    1,
    1,
    4,
    1,
    5,
    1,
    1,
    1,
    6,
    1,
    1,
    1,
    5,
    1,
    5,
    1,
    5,
    1
  )

  # mandatory fields, fast read
  dat <- readr::read_fwf(
    tmp,
    readr::fwf_widths(column_widths),
    col_types = "ccciiiiicccccccicciccccccccccccccc"
  )

  # additional fields, variable length, need to read eveything
  add <- readLines(tmp)

  # Remove select columns from data frame
  dat <- dat[, c(2:8, 10:11, 13, 16, 19, 21, 25, 29, 31, 33)]

  # Apply new names to the data frame columns
  names(dat) <- c(
    "usaf",
    "wban",
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "latitude",
    "longitude",
    "elev",
    "wd",
    "ws",
    "ceil_hgt",
    "visibility",
    "air_temp",
    "dew_point",
    "atmos_pres"
  )

  # Correct the latitude values
  dat$latitude <- as.numeric(dat$latitude) / 1000

  # Correct the longitude values
  dat$longitude <- as.numeric(dat$longitude) / 1000

  # Correct the elevation values
  dat$elev <- as.numeric(dat$elev)

  # Correct the wind direction values
  dat$wd <-
    ifelse(dat$wd == 999, NA, dat$wd)

  # Correct the wind speed values
  dat$ws <-
    ifelse(dat$ws == 9999, NA, dat$ws / 10)

  # Correct the temperature values
  dat$air_temp <- as.numeric(dat$air_temp)
  dat$air_temp <- ifelse(dat$air_temp == 9999, NA, dat$air_temp / 10)

  # Correct the visibility values
  dat$visibility <- as.numeric(dat$visibility)
  dat$visibility <- ifelse(
    dat$visibility %in% c(9999, 999999),
    NA,
    dat$visibility
  )

  # Correct the dew point values
  dat$dew_point <- as.numeric(dat$dew_point)
  dat$dew_point <- ifelse(dat$dew_point == 9999, NA, dat$dew_point / 10)

  # Correct the atmospheric pressure values
  dat$atmos_pres <- as.numeric(dat$atmos_pres)
  dat$atmos_pres <- ifelse(dat$atmos_pres == 99999, NA, dat$atmos_pres / 10)

  # Correct the ceiling height values
  dat$ceil_hgt <- as.numeric(dat$ceil_hgt)
  dat$ceil_hgt <- ifelse(dat$ceil_hgt == 99999, NA, dat$ceil_hgt)

  # relative humidity - general formula based on T and dew point
  dat$RH <- 100 *
    ((112 - 0.1 * dat$air_temp + dat$dew_point) /
      (112 + 0.9 * dat$air_temp))^8

  dat$date <- ISOdatetime(
    dat$year,
    dat$month,
    dat$day,
    dat$hour,
    dat$minute,
    0,
    tz = "GMT"
  )

  # drop date components that are not needed
  dat <- dplyr::select(
    dat,
    -dplyr::all_of(c("year", "month", "day", "hour", "minute"))
  )

  # process the additional data separately
  dat <- procAdditFwf(add, dat, precip, PWC)

  # for cloud cover, make new 'cl' max of 3 cloud layers
  dat$cl <- pmax(dat$cl_1, dat$cl_2, dat$cl_3, na.rm = TRUE)

  # additional condition, when ceil_height = 22000 and cl_1 is NA, assume no cloud
  dat <- dat |>
    dplyr::mutate(
      cl = ifelse(
        (is.na(.data$cl_1) & .data$ceil_hgt == 22000),
        0,
        .data$cl
      ),
      cl_1 = ifelse(
        (is.na(.data$cl_1) & .data$ceil_hgt == 22000),
        0,
        .data$cl_1
      )
    )

  # select the variables we want
  dat <- dat[
    names(dat) %in%
      c(
        "date",
        "usaf",
        "wban",
        "station",
        "ws",
        "wd",
        "air_temp",
        "atmos_pres",
        "visibility",
        "dew_point",
        "RH",
        "ceil_hgt",
        "latitude",
        "longitude",
        "elev",
        "cl_1",
        "cl_2",
        "cl_3",
        "cl",
        "cl_1_height",
        "cl_2_height",
        "cl_3_height",
        "pwc",
        "precip_12",
        "precip_6"
      )
  ]

  # present weather is character and cannot be averaged, take first
  if ("pwc" %in% names(dat) && hourly) {
    pwc <- dat[c("date", "pwc")]
    pwc$date2 <- format(pwc$date, "%Y-%m-%d %H") # nearest hour
    tmp <- pwc[which(!duplicated(pwc$date2)), ]
    dates <- as.POSIXct(paste0(unique(pwc$date2), ":00:00"), tz = "GMT")

    pwc <- data.frame(date = dates, pwc = tmp$pwc)
    PWC <- TRUE
  }

  # average to hourly
  if (hourly) {
    dat <- worldmet_time_average(
      dat,
      avg.time = "hour",
      type = c("usaf", "wban")
    )
  }

  # add precipitation
  if (precip) {
    # spread out precipitation across each hour
    # met data gives 12 hour total and every other 6 hour total

    # only do this if precipitation exists
    if (all(c("precip_6", "precip_12") %in% names(dat))) {
      # make new precip variable
      dat$precip <- NA

      # id where there is 6 hour data
      id <- which(!is.na(dat$precip_6))
      id <- id[id < (nrow(dat) - 6)] # make sure we don't run off end

      # calculate new 6 hour based on 12 hr total - 6 hr total
      dat$precip_6[id + 6] <- dat$precip_12[id + 6] - dat$precip_6[id]

      # ids for new 6 hr totals
      id <- which(!is.na(dat$precip_6))
      id <- id[id > 6]

      # Divide 6 hour total over each of 6 hours
      for (i in seq_along(id)) {
        dat$precip[(id[i] - 5):id[i]] <- dat$precip_6[id[i]] / 6
      }
    }
  }

  # return other meta data
  meta <- getMeta(returnMap = FALSE, plot = FALSE, end.year = "all")
  info <- meta[meta$code == code, ]

  dat$station <- as.character(info$station)
  dat$latitude <- info$latitude
  dat$longitude <- info$longitude
  dat$elev <- info$`elev(m)`
  dat$code <- code

  # rearrange columns, one to move to front
  move <- c("date", "code", "station")
  dat <- dat[c(move, setdiff(names(dat), move))]

  # replace NaN with NA
  dat <- dplyr::mutate(
    dat,
    dplyr::across(dplyr::everything(), \(x) dplyr::if_else(is.nan(x), NA, x))
  )

  # reorder to match delimited data
  dat <-
    dplyr::relocate(
      dat,
      dplyr::any_of(
        c(
          "code",
          "station",
          "date",
          "latitude",
          "longitude",
          "elev",
          "ws",
          "wd",
          "air_temp",
          "atmos_pres",
          "visibility",
          "dew_point",
          "RH",
          "ceil_hgt",
          "cl_1",
          "cl_2",
          "cl_3",
          "cl",
          "cl_1_height",
          "cl_2_height",
          "cl_3_height",
          "precip_12",
          "precip_6",
          "pwc",
          "precip"
        )
      )
    )

  return(dplyr::tibble(dat))
}

#' Deprecated ISD access functions
#'
#' @description `r lifecycle::badge('deprecated')`
#'
#' This function is part of an old `worldmet` API. Please use the following
#' alternatives:
#' - [getMeta()] -> [import_isd_stations()]
#' - [getMetaLive()] -> [import_isd_stations_live()]
#' - [importNOAA()] -> [import_isd_hourly()]
#' - [importNOAAlite()] -> [import_isd_lite()]
#' - the `path` argument -> [write_met()]
#'
#' @inheritParams import_isd_stations
#' @inheritParams import_isd_stations_live
#' @inheritParams import_isd_hourly
#'
#' @param lat,lon,n Decimal latitude (`lat`) and longitude (`lon`) (or other Y/X
#'   coordinate if using a different `crs`). If provided, the `n` closest ISD
#'   stations to this coordinate will be returned.
#' @param end.year To help filter sites based on how recent the available data
#'   are. `end.year` can be "current", "any" or a numeric year such as 2016, or
#'   a range of years e.g. 1990:2016 (which would select any site that had an
#'   end date in that range. **By default only sites that have some data for the
#'   current year are returned**.
#' @param plot If `TRUE` will plot sites on an interactive leaflet map.
#' @param returnMap Should the leaflet map be returned instead of the meta data?
#'   Default is `FALSE`.
#'
#' @return A data frame is returned with all available meta data, mostly
#'   importantly including a `code` that can be supplied to [importNOAA()]. If
#'   latitude and longitude searches are made an approximate distance, `dist` in
#'   km is also returned.
#'
#' @rdname deprecated-isd
#' @export
getMeta <- function(
  site = NULL,
  lat = NULL,
  lon = NULL,
  crs = 4326,
  country = NULL,
  state = NULL,
  n = 10,
  end.year = "current",
  provider = c(
    "Street Map" = "CartoDB.Voyager",
    "Satellite" = "Esri.WorldImagery"
  ),
  plot = TRUE,
  returnMap = FALSE
) {
  if (plot || returnMap) {
    m <- import_isd_stations(
      site = site,
      lat = lat,
      lng = lon,
      crs = crs,
      country = country,
      state = state,
      n_max = n,
      end_year = end.year,
      provider = provider,
      return = "map"
    )
  }

  if (plot) {
    print(m)
  }

  if (returnMap) {
    return(m)
  } else {
    dat <-
      import_isd_stations(
        site = site,
        lat = lat,
        lng = lon,
        crs = crs,
        country = country,
        state = state,
        n_max = n,
        end_year = end.year,
        provider = provider,
        return = "table"
      )
    return(dat)
  }
}

#' @rdname deprecated-isd
#' @export
getMetaLive <- function(...) {
  import_isd_stations_live(...)
}

#' @param path If a file path is provided, the data are saved as an rds file at
#'   the chosen location e.g.  `path = "C:/Users/David"`. Files are saved by
#'   year and site.
#' @param n.cores No longer recommended; please set [mirai::daemons()] in your R
#'   session. This argument is provided for back compatibility, and is passed to
#'   the `n` argument of [mirai::daemons()] on behalf of the user. Any set
#'   daemons will be reset once the function completes. Default is `NULL`, which
#'   means no parallelism. `n.cores = 1L` is equivalent to `n.cores = NULL`.
#' @rdname deprecated-isd
#' @export
importNOAA <- function(
  code = "037720-99999",
  year = 2014,
  hourly = TRUE,
  source = c("delim", "fwf"),
  quiet = FALSE,
  path = NA,
  n.cores = NULL
) {
  if (!is.null(n.cores)) {
    if (!rlang::is_integerish(n.cores)) {
      cli::cli_abort(
        "{.field n.cores} should be an integer. You have provided {.type {n.cores}}."
      )
    }
    if (n.cores > 1L) {
      if (mirai::daemons_set()) {
        cli::cli_warn(
          "{.fun mirai::daemons} have already been set. Ignoring {.field n.cores}."
        )
      } else {
        on.exit(mirai::daemons(0))
        cli::cli_warn(
          c(
            "!" = "Using {.fun mirai::daemons} to set multiple workers. These will be reset at the end of the function call.",
            "i" = "It is now preferred that users use {.fun mirai::daemons} directly over the use of {.field n.cores}.",
            "i" = "See {.url https://openair-project.github.io/worldmet/articles/worldmet.html} for an example."
          ),
          .frequency = "regularly",
          .frequency_id = "worldmet_mirai"
        )
        mirai::daemons(n.cores)
      }
    }
  }

  dat <-
    import_isd_hourly(
      code = code,
      year = year,
      hourly = hourly,
      source = source,
      progress = !quiet,
      quiet = quiet
    )

  if (is.null(dat)) {
    return(NULL)
  }

  if (!is.na(path)) {
    write_met(
      x = dat,
      path = path,
      ext = "rds",
      suffix = "",
      progress = FALSE
    )
  }

  return(dat)
}

#' @rdname deprecated-isd
#' @export
importNOAAlite <- function(
  code = "037720-99999",
  year = 2025,
  quiet = FALSE,
  path = NA
) {
  dat <-
    import_isd_lite(
      code = code,
      year = year,
      progress = !quiet,
      quiet = quiet
    )

  if (is.null(dat)) {
    return(NULL)
  }

  if (!is.na(path)) {
    write_met(
      x = dat,
      path = path,
      ext = "rds",
      suffix = "_lite",
      progress = FALSE
    )
  }

  return(dat)
}

#' Deprecated data functions
#'
#' @description
#' `r lifecycle::badge('deprecated')`
#'
#' This function is part of an old `worldmet` API. Please use the following
#' alternatives:
#' - [exportADMS()] -> [write_adms()]
#' - the `path` argument -> [write_met()]
#'
#' @param dat A data frame imported by [importNOAA()].
#' @param out A file name for the ADMS file. The file is written to the working
#'   directory by default.
#' @param interp Should interpolation of missing values be undertaken? If `TRUE`
#'   linear interpolation is carried out for gaps of up to and including
#'   `maxgap`.
#' @param maxgap The maximum gap in hours that should be interpolated where
#'   there are missing data when `interp = TRUE.` Data with gaps more than
#'   `maxgap` are left as missing.
#'
#' @rdname deprecated-data
#' @export
exportADMS <- function(
  dat,
  out = "./ADMS_met.MET",
  interp = FALSE,
  maxgap = 2
) {
  write_adms(
    x = dat,
    file = out,
    interp = interp,
    max_gap = maxgap
  )
}

#' Function to alert that the ISD is going away
#' @noRd
warn_isd <- function(fn = cli::cli_warn) {
  fn(
    c(
      "!" = "The integrated surface database has been deprecated by NOAA, and data is now only available until 2025.",
      "i" = "Please consider using {.fun worldmet::import_ghcn_stations} and {.fun worldmet::import_ghcn_hourly} to access data from the new Global Historical Climatology Network."
    ),
    .frequency = "regularly",
    .frequency_id = "isd-deprecated"
  )
}

#' Import data from the Global Historical Climatology monthly (GHCNm) database
#'
#' This function is a convenient way to access the monthly summaries of the
#' GHCN. Monthly average temperature is available via
#' [import_ghcn_monthly_temp()] and monthly precipitation via
#' [import_ghcn_monthly_prcp()]. Note that these functions can take a few
#' minutes to run, and parallelism is only enabled for precipitation data.
#'
#' @inheritSection import_ghcn_hourly Parallel Processing
#'
#' @inheritParams import_ghcn_hourly
#'
#' @param table Either `"inventory"`, `"data"`, or both. The tables to read and
#'   return in the output list.
#'
#' @param dataset For [import_ghcn_monthly_temp()]. One of the below options.
#'   More information is available at
#'   <https://www.ncei.noaa.gov/pub/data/ghcn/v4/readme.txt>.
#'
#' - `"qcu"`: Quality Control, Unadjusted
#'
#' - `"qcf"`: Quality Control, Adjusted, using the Pairwise Homogeneity
#'   Algorithm.
#'
#' - `"qfe"`: Quality Control, Adjusted, Estimated using the Pairwise
#'   Homogeneity Algorithm. Only the years 1961-2010 are provided. This is to
#'   help maximize station coverage when calculating normals.
#'
#' @param station For [import_ghcn_monthly_prcp()]. The specific stations to
#'   import monthly precipitation data for.
#'
#' @param progress For [import_ghcn_monthly_prcp()]. Show a progress bar when
#'   importing many stations? Defaults to `TRUE` in interactive R sessions.
#'   Passed to `.progress` in [purrr::map()].
#'
#' @author Jack Davison
#' @family GHCN functions
#'
#' @rdname import_ghcn_monthly
#' @return a list of [tibbles][tibble::tibble-package]
#' @export
import_ghcn_monthly_temp <- function(
  table = c("inventory", "data"),
  dataset = c("qcu", "qcf", "qfe")
) {
  # id temporary directory to download into
  tdir <- tempdir()
  folder <- paste0(
    "ghcnm_",
    gsub("\\.", "_", as.character(as.numeric(Sys.time())))
  )
  dir.create(file.path(tdir, folder))

  # ensure dataset is correct
  dataset <- rlang::arg_match(dataset, c("qcu", "qcf", "qfe"))
  table <- rlang::arg_match(table, c("inventory", "data"), multiple = TRUE)

  # download file into temporary dir
  utils::download.file(
    paste0(
      "https://www.ncei.noaa.gov/pub/data/ghcn/v4/ghcnm.tavg.latest.",
      dataset,
      ".tar.gz"
    ),
    destfile = file.path(
      tdir,
      folder,
      "out.tar.gz"
    ),
    mode = "wb"
  )

  # untar it
  utils::untar(
    file.path(tdir, folder, "out.tar.gz"),
    exdir = file.path(tdir, folder)
  )

  # get the file that's been extracted
  files <- dir(file.path(tdir, folder), full.names = TRUE)
  files <- files[!grepl("out.tar.gz", files)]
  files <- dir(files, full.names = TRUE)

  out <- list()

  if ("data" %in% table) {
    # get month names for columns
    monthnames <- c()
    for (i in month.abb) {
      monthnames <- c(
        monthnames,
        paste0(i, c("_value", "_dmf", "_qcf", "_dsf"))
      )
    }
    monthnames <- tolower(monthnames)

    # read data file
    dat <-
      readr::read_fwf(
        files[grepl("dat", files)],
        col_positions = readr::fwf_widths(c(
          11,
          4,
          4,
          rep(c(5, 1, 1, 1), 12)
        )),
        na = "-9999",
        show_col_types = FALSE
      ) |>
      stats::setNames(
        c(
          "station_id",
          "year",
          "element",
          monthnames
        )
      ) |>
      dplyr::mutate(
        dplyr::across(dplyr::ends_with("_value"), \(x) x / 100),
        dplyr::across(dplyr::ends_with("_dmf|_qcf|_dsf"), as.character)
      )

    out <- append(out, list(data = dat))
  }

  if ("inventory" %in% table) {
    # read inventory file
    inv <-
      readr::read_fwf(
        files[grepl("inv", files)],
        col_positions = readr::fwf_cols(
          station_id = c(1, 11),
          lat = c(13, 20),
          lng = c(22, 30),
          elev = c(32, 37),
          station_name = c(39, 68)
        ),
        show_col_types = FALSE
      )

    out <- append(out, list(inventory = inv))
  }

  return(out)
}

#' @rdname import_ghcn_monthly
#' @export
import_ghcn_monthly_prcp <- function(
  station = NULL,
  year = NULL,
  table = c("inventory", "data"),
  progress = rlang::is_interactive()
) {
  table <- rlang::arg_match(table, c("inventory", "data"), multiple = TRUE)

  out <- list()

  if ("data" %in% table) {
    # read data
    data <-
      purrr::map(
        .x = station,
        .f = purrr::in_parallel(
          .f = \(x) {
            data <- try(
              suppressWarnings(
                readr::read_csv(
                  paste0(
                    "https://www.ncei.noaa.gov/data/ghcnm/v4/precipitation/access/",
                    x,
                    ".csv"
                  ),
                  col_names = c(
                    "station_id",
                    "station_name",
                    "lat",
                    "lng",
                    "elev",
                    "yearmon",
                    "precip",
                    "meas_flag",
                    "qc_flag",
                    "src_flag",
                    "src_index"
                  ),
                  show_col_types = FALSE,
                  progress = FALSE
                )
              ),
              silent = TRUE
            )

            if (class(data)[1] == "try-error") {
              return(NULL)
            }

            return(data)
          }
        ),
        .progress = progress
      ) |>
      dplyr::bind_rows()

    if (is.null(data) || nrow(data) == 0L) {
      cli::cli_warn("No data has been returned.")
    } else {
      data <- data |>
        dplyr::mutate(
          year = substr(.data$yearmon, 1, 4),
          month = substr(.data$yearmon, 5, 6),
          date = ISOdate(.data$year, .data$month, 1, hour = 0) |> as.Date(),
          precip = .data$precip * 0.1,
          .keep = "unused"
        ) |>
        dplyr::relocate("date", "year", "month", .before = "precip")
    }

    out <- append(out, list(data = data))
  }

  if ("inventory" %in% table) {
    # get inventory
    inv_lines <- readr::read_lines(
      "https://www.ncei.noaa.gov/data/ghcnm/v4/precipitation/doc/ghcn-m_v4_prcp_inventory.txt",
      progress = FALSE
    )

    inv <-
      readr::read_fwf(
        I(inv_lines),
        col_positions = readr::fwf_cols(
          station_id = c(1, 11),
          lat = c(13, 20),
          lng = c(22, 30),
          elev = c(32, 37),
          state = c(39, 40),
          station_name = c(42, 79),
          wmo_id = c(81, 85),
          start = c(87, 90),
          end = c(92, 95)
        ),
        na = c("-999.9"),
        show_col_types = FALSE
      )

    out <- append(out, list(inventory = inv))
  }

  if (!is.null(year)) {
    data <- dplyr::filter(
      data,
      as.numeric(format(.data$date, "%Y")) %in% year
    )

    if (nrow(data) == 0L) {
      cli::cli_warn("No data has been returned. Check {.arg year}.")
      return(NULL)
    }
  }

  return(out)
}

#' Codes for weather types
#'
#' This data frame consists of the weather description codes used in the ISD. It
#' is not of general use to most users.
#'
#' \describe{
#'   \item{pwc}{Weather code, which can be merged with the `pwc` column in [importNOAA()] datasets.}
#'   \item{description}{Description associated with the weather codes.}
#' }
#' @examples
#' weatherCodes
"weatherCodes"

#' @keywords internal
#'
#' @details This package contains functions to import surface meteorological
#'   data from over 30,000 sites around the world. These data are curated by
#'   NOAA as part of the Global Historical Climate Network (GHCN), which
#'   replaced the Integrated Surface Database (ISD) in 2025.
#'
#'   If you access these data using the `worldmet` package please give full
#'   acknowledgement to NOAA. Users should also take a note of the usage
#'   restrictions.
#'
#'   These data work well with the `openair` package that has been developed to
#'   analyse air pollution data.
#'
#' @references
#'
#' For general information about the GHCNh, see
#' <https://www.ncei.noaa.gov/products/global-historical-climatology-network-hourly>.
#'
#' For general information about the legacy ISD service, see
#' <https://www.ncei.noaa.gov/products/land-based-station/integrated-surface-database>.
#'
#' @seealso See <https://github.com/openair-project/openair> for information on
#'   the related `openair` package.
"_PACKAGE"
## usethis namespace: start
#' @importFrom carrier crate
#' @importFrom lifecycle deprecated
#' @importFrom mirai daemons
#' @importFrom rlang %||%
#' @importFrom rlang .data
#' @importFrom tibble tibble
## usethis namespace: end
NULL

#' Export a meteorological data frame in files, chunked by site and year
#'
#' Writes data returned by any of [import_isd_hourly()], [import_ghcn_hourly()],
#' or [import_ghcn_daily()] to a file. Each station and year in the data is
#' written to a separate file.
#'
#' @param x A data frame imported by [import_isd_hourly()],
#'   [import_ghcn_hourly()], or [import_ghcn_daily()].
#'
#' @param path The path to a directory to save each file. By default, this is
#'   the working directory.
#'
#' @param ext The file type to use when saving the data. Can be `"rds"`,
#'   `"delim"` or `"parquet"`. Note that `"parquet"` requires the `arrow`
#'   package.
#'
#' @param delim Delimiter used to separate values when `ext = "delim"`. Must be
#'   a single character. Defaults to being comma-delimited (`","`).
#'
#' @param suffix An additional suffix to append to file names. Useful examples
#'   could be `"_isd"`, `"_hourly"`, `"_lite"`, and so on.
#'
#' @param progress Show a progress bar when writing many stations/years?
#'   Defaults to `TRUE` in interactive R sessions. Passed to `.progress` in
#'   [purrr::walk()].
#'
#' @return `write_met()` returns `path` invisibly.
#' @family Met writing functions
#' @export
#' @examples
#' \dontrun{
#' # import some data then export it
#' dat <- import_isd_hourly(year = 2012)
#' write_met(dat)
#' }
write_met <- function(
  x,
  path = ".",
  ext = c("rds", "delim", "parquet"),
  delim = ",",
  suffix = "",
  progress = rlang::is_interactive()
) {
  ext <- rlang::arg_match(ext, c("rds", "delim", "parquet"))

  # error if path doesn't exist
  if (!dir.exists(path)) {
    cli::cli_abort("Directory does not exist; file not saved.", call. = FALSE)
  }

  # find the ID column for the data
  if ("id" %in% names(x)) {
    id_col <- "id"
  } else if ("station_id" %in% names(x)) {
    id_col <- "station_id"
  } else if ("code" %in% names(x)) {
    id_col <- "code"
  } else {
    cli::cli_abort(
      c(
        "x" = "Data not recognised.",
        "i" = "Did you access it via {.fun worldmet::import_isd_hourly}, {.fun worldmet::import_ghcn_hourly}, or {.fun worldmet::import_ghcn_daily}?"
      )
    )
  }

  # save as year / site files
  write_met_helper <- function(x) {
    filepath <- paste0(
      path,
      "/",
      unique(x[[id_col]]),
      "_",
      unique(x$year),
      suffix,
      ".",
      ext
    )

    if (ext == "rds") {
      saveRDS(x, filepath)
    }

    if (ext == "csv") {
      readr::write_delim(x, filepath)
    }

    if (ext == "parquet") {
      rlang::check_installed("arrow")
      arrow::write_parquet(x, filepath)
    }
  }

  x |>
    dplyr::mutate(year = format(.data$date, "%Y")) |>
    (\(df) split(df, df[c(id_col, "year")]))() |>
    purrr::walk(.f = write_met_helper, .progress = progress)

  invisible(path)
}

#' Import data from the Global Historical Climatology daily (GHCNd) database
#'
#' This function flexibly accesses meteorological data from the GHCNd database.
#' Users can provide any of stations, and control whether attribute codes are
#' returned with the data.
#'
#' @section Data Definition:
#'
#'   The core elements in the GHCNd are:
#'
#'   - **PRCP**: Precipitation (mm)
#'   - **SNOW**: Snowfall (mm)
#'   - **SNWD**: Snow depth (mm)
#'   - **TMAX**: Maximum temperature (degrees C)
#'   - **TMIN**: Minimum temperature (degrees C)
#'
#'   Other elements which may appear are:
#'
#'   - **ACMC**: Average cloudiness midnight to midnight from 30-second ceilometer data (percent)
#'   - **ACMH**: Average cloudiness midnight to midnight from manual observations (percent)
#'   - **ACSC**: Average cloudiness sunrise to sunset from 30-second ceilometer data (percent)
#'   - **ACSH**: Average cloudiness sunrise to sunset from manual observations (percent)
#'   - **ADPT**: Average Dew Point Temperature for the day (degrees C)
#'   - **ASLP**: Average Sea Level Pressure for the day (hPa)
#'   - **ASTP**: Average Station Level Pressure for the day (hPa)
#'   - **AWBT**: Average Wet Bulb Temperature for the day (degrees C)
#'   - **AWDR**: Average daily wind direction (degrees)
#'   - **AWND**: Average daily wind speed (meters per second)
#'   - **DAEV**: Number of days included in the multiday evaporation total (MDEV)
#'   - **DAPR**: Number of days included in the multiday precipitation total (MDPR)
#'   - **DASF**: Number of days included in the multiday snowfall total (MDSF)
#'   - **DATN**: Number of days included in the multiday minimum temperature (MDTN)
#'   - **DATX**: Number of days included in the multiday maximum temperature (MDTX)
#'   - **DAWM**: Number of days included in the multiday wind movement (MDWM)
#'   - **DWPR**: Number of days with non-zero precipitation included in multiday precipitation total (MDPR)
#'   - **EVAP**: Evaporation of water from evaporation pan (mm)
#'   - **FMTM**: Time of fastest mile or fastest 1-minute wind (hours and minutes, i.e., HHMM)
#'   - **FRGB**: Base of frozen ground layer (cm)
#'   - **FRGT**: Top of frozen ground layer (cm)
#'   - **FRTH**: Thickness of frozen ground layer (cm)
#'   - **GAHT**: Difference between river and gauge height (cm)
#'   - **MDEV**: Multiday evaporation total (mm; use with **DAEV**)
#'   - **MDPR**: Multiday precipitation total (mm; use with **DAPR** and **DWPR**, if available)
#'   - **MDSF**: Multiday snowfall total
#'   - **MDTN**: Multiday minimum temperature (degrees C; use with **DATN**)
#'   - **MDTX**: Multiday maximum temperature (degrees C; use with **DATX**)
#'   - **MDWM**: Multiday wind movement (km)
#'   - **MNPN**: Daily minimum temperature of water in an evaporation pan (degrees C)
#'   - **MXPN**: Daily maximum temperature of water in an evaporation pan (degrees C)
#'   - **PGTM**: Peak gust time (hours and minutes, i.e., HHMM)
#'   - **PSUN**: Daily percent of possible sunshine (percent)
#'   - **RHAV**: Average relative humidity for the day (percent)
#'   - **RHMN**: Minimum relative humidity for the day (percent)
#'   - **RHMX**: Maximum relative humidity for the day (percent)
#'   - **SN$#**: Minimum soil temperature (degrees C) where $ corresponds to a code for ground cover and # corresponds to a code for soil depth. Ground cover codes: 0=unknown, 1=grass, 2=fallow, 3=bare ground, 4=brome grass, 5=sod, 6=straw mulch, 7=grass muck, 8=bare muck. Depth codes: 1=5cm, 2=10cm, 3=20cm, 4=50cm, 5=100cm, 6=150cm, 7=180cm
#'   - **SX$#**: Maximum soil temperature (degrees C) where $ corresponds to a code for ground cover and # corresponds to a code for soil depth. See **SN$#** for ground cover and depth codes
#'   - **TAXN**: Average daily temperature computed as `(TMAX+TMIN)/2.0` (degrees C)
#'   - **TAVG**: Average daily temperature (degrees C). Note that TAVG from source 'S' corresponds to an average of hourly readings for the period ending at 2400 UTC rather than local midnight or other Local Standard Time according to a specific Met Service's protocol. For sources other than 'S' TAVG is computed in a variety of ways including traditional fixed hours of the day whereas TAXN is solely computed as (TMAX+TMIN)/2.0
#'   - **THIC**: Thickness of ice on water (mm)
#'   - **TOBS**: Temperature at the time of observation (degrees C)
#'   - **TSUN**: Daily total sunshine (minutes)
#'   - **WDF1**: Direction of fastest 1-minute wind (degrees)
#'   - **WDF2**: Direction of fastest 2-minute wind (degrees)
#'   - **WDF5**: Direction of fastest 5-second wind (degrees)
#'   - **WDFG**: Direction of peak wind gust (degrees)
#'   - **WDFI**: Direction of highest instantaneous wind (degrees)
#'   - **WDFM**: Fastest mile wind direction (degrees)
#'   - **WDMV**: 24-hour wind movement (km)
#'   - **WESD**: Water equivalent of snow on the ground (mm)
#'   - **WESF**: Water equivalent of snowfall (mm)
#'   - **WSF1**: Fastest 1-minute wind speed (meters per second)
#'   - **WSF2**: Fastest 2-minute wind speed (meters per second)
#'   - **WSF5**: Fastest 5-second wind speed (meters per second)
#'   - **WSFG**: Peak gust wind speed (meters per second)
#'   - **WSFI**: Highest instantaneous wind speed (meters per second)
#'   - **WSFM**: Fastest mile wind speed (meters per second)
#'
#'   There can be any number of weather types (**WT$$**)
#'
#'   - **WT01**: Fog, ice fog, or freezing fog (may include heavy fog)
#'   - **WT02**: Heavy fog or heaving freezing fog (not always distinguished from fog)
#'   - **WT03**: Thunder
#'   - **WT04**: Ice pellets, sleet, snow pellets, or small hail
#'   - **WT05**: Hail (may include small hail)
#'   - **WT06**: Glaze or rime
#'   - **WT07**: Dust, volcanic ash, blowing dust, blowing sand, or blowing obstruction
#'   - **WT08**: Smoke or haze
#'   - **WT09**: Blowing or drifting snow
#'   - **WT10**: Tornado, waterspout, or funnel cloud
#'   - **WT11**: High or damaging winds
#'   - **WT12**: Blowing spray
#'   - **WT13**: Mist
#'   - **WT14**: Drizzle
#'   - **WT15**: Freezing drizzle
#'   - **WT16**: Rain (may include freezing rain, drizzle, and freezing drizzle)
#'   - **WT17**: Freezing rain
#'   - **WT18**: Snow, snow pellets, snow grains, or ice crystals
#'   - **WT19**: Unknown source of precipitation
#'   - **WT21**: Ground fog
#'   - **WT22**: Ice fog or freezing fog
#'
#'   There can also be any number of 'weather in the vicinity' columns
#'   (**WV$$**)
#'
#'   - **WV01**: Fog, ice fog, or freezing fog (may include heavy fog)
#'   - **WV03**: Thunder
#'   - **WV07**: Ash, dust, sand, or other blowing obstruction
#'   - **WV18**: Snow or ice crystals
#'   - **WV20**: Rain or snow shower
#'
#' @inheritSection import_ghcn_hourly Parallel Processing
#' @inheritParams import_ghcn_hourly
#'
#' @param source The data format for the GHCNd. Currently only `"csv"` is
#'   supported. This argument is included for future use.
#'
#' @param codes When `append_codes` is `TRUE`, which codes should be appended to
#'   the dataframe? Any combination of `"measurement_flag"`, `"quality_flag"`,
#'   and/or `"source_flag"`.
#'
#' @author Jack Davison
#' @family GHCN functions
#'
#' @return a [tibble][tibble::tibble-package]
#' @export
import_ghcn_daily <- function(
  station,
  year = NULL,
  source = c("csv"),
  extra = FALSE,
  append_codes = FALSE,
  codes = c(
    "measurement_flag",
    "quality_flag",
    "source_flag"
  ),
  progress = rlang::is_interactive()
) {
  # check inputs
  source <- rlang::arg_match(source, c("csv"), multiple = FALSE)
  codes <- rlang::arg_match(
    codes,
    c(
      "measurement_flag",
      "quality_flag",
      "source_flag"
    ),
    multiple = TRUE
  )

  # import data
  data <-
    purrr::map(
      .x = station,
      .f = purrr::in_parallel(\(x) {
        data <- try(
          suppressWarnings(
            readr::read_csv(
              paste0(
                "https://www.ncei.noaa.gov/data/global-historical-climatology-network-daily/access/",
                x,
                ".csv"
              ),
              name_repair = tolower,
              show_col_types = FALSE,
              progress = FALSE
            )
          ),
          silent = TRUE
        )

        if (class(data)[1] == "try-error") {
          return(NULL)
        }

        return(data)
      }),
      .progress = progress
    ) |>
    dplyr::bind_rows()

  # if null, leave early
  if (is.null(data) || nrow(data) == 0L) {
    cli::cli_warn("No data has been returned.")
    return(NULL)
  }

  # sort out names
  data <-
    data |>
    dplyr::rename(
      station_id = "station",
      station_name = "name",
      lat = "latitude",
      lng = "longitude",
      elev = "elevation"
    ) |>
    dplyr::relocate(
      "station_name",
      .after = "station_id"
    )

  if (!append_codes) {
    data <- dplyr::select(data, -dplyr::ends_with("attributes"))
  } else {
    data <- tidyr::separate_wider_delim(
      data,
      dplyr::ends_with("_attributes"),
      delim = ",",
      names_sep = "_",
      names = c("mf", "qf", "sf")
    ) |>
      dplyr::rename_with(\(x) gsub("_attributes_", "_", x))

    if (!"measurement_flag" %in% codes) {
      data <- dplyr::select(data, -dplyr::ends_with("_mf"))
    }
    if (!"quality_flag" %in% codes) {
      data <- dplyr::select(data, -dplyr::ends_with("_qf"))
    }
    if (!"source_flag" %in% codes) {
      data <- dplyr::select(data, -dplyr::ends_with("_sf"))
    }
  }

  # deal with "tenths of" from documentation
  for (i in c(
    "prcp",
    "tmax",
    "tmin",
    "adpt",
    "aslp",
    "astp",
    "awbt",
    "awnd",
    "evap",
    "mdev",
    "mdpr",
    "mdtn",
    "mdtx",
    "mnpn",
    "mxpn",
    "taxn",
    "tavg",
    "thic",
    "tobs",
    "wesd",
    "wesf",
    "wsf1",
    "wsf2",
    "wsf5",
    "wsfg",
    "wsfi",
    "wsfm"
  )) {
    if (i %in% names(data)) {
      data[i] <- data[[i]] / 10
    }
  }

  # soil could be lots of things
  soil_names <- names(data)[
    startsWith(names(data), "sn") &
      nchar(names(data)) == 4 &
      !names(data) %in% c("snwd", "snow")
  ]
  if (length(soil_names) > 1) {
    for (i in soil_names) {
      data[i] <- data[[i]] / 10
    }
  }

  sx_names <- names(data)[
    startsWith(names(data), "sx")
  ]
  if (length(sx_names) > 1) {
    for (i in sx_names) {
      data[i] <- data[[i]] / 10
    }
  }

  if (!extra) {
    data <- dplyr::select(
      data,
      dplyr::any_of(c(
        "station_id",
        "station_name",
        "date",
        "lat",
        "lng",
        "elev",
        "prcp",
        "snow",
        "snwd",
        "tavg",
        "tmax",
        "tmin"
      ))
    )
  }

  if (!is.null(year)) {
    data <- dplyr::filter(
      data,
      as.numeric(format(.data$date, "%Y")) %in% year
    )

    if (nrow(data) == 0L) {
      cli::cli_warn("No data has been returned. Check {.arg year}.")
      return(NULL)
    }
  }

  return(data)
}

#' Import station metadata for the Integrated Surface Databse
#'
#' This function is primarily used to find a site code that can be used to
#' access data using [import_isd_hourly()]. Sites searches of approximately
#' 30,000 sites can be carried out based on the site name and based on the
#' nearest locations based on user-supplied latitude and longitude.
#'
#' @inheritParams import_ghcn_stations
#'
#' @param site A site name search string e.g. `site = "heathrow"`. The search
#'   strings and be partial and can be upper or lower case e.g. `site =
#'   "HEATHR"`.
#'
#' @param country The country code. This is a two letter code. For a full
#'   listing see <https://www.ncei.noaa.gov/pub/data/noaa/isd-history.csv>.
#'
#' @param state The state code. This is a two letter code.
#'
#' @param end_year To help filter sites based on how recent the available data
#'   are. `end_year` can be "current", "any" or a numeric year such as 2016, or
#'   a range of years e.g. 1990:2016 (which would select any site that had an
#'   end date in that range. **By default only sites that have some data for the
#'   current year are returned**.
#'
#' @return A data frame is returned with all available meta data, mostly
#'   importantly including a `code` that can be supplied to [importNOAA()]. If
#'   latitude and longitude searches are made an approximate distance, `dist` in
#'   km is also returned.
#'
#' @export
#' @author David Carslaw
#' @family NOAA ISD functions
#' @examples
#' \dontrun{
#' ## search for sites with name beijing
#' getMeta(site = "beijing")
#' }
#'
#' \dontrun{
#' ## search for near a specified lat/lng - near Beijing airport
#' ## returns 'n_max' nearest by default
#' getMeta(lat = 40, lng = 116.9)
#' }
import_isd_stations <- function(
  site = NULL,
  country = NULL,
  state = NULL,
  lat = NULL,
  lng = NULL,
  crs = 4326,
  n_max = 10,
  end_year = "current",
  provider = c(
    "Street Map" = "CartoDB.Voyager",
    "Satellite" = "Esri.WorldImagery"
  ),
  return = c("table", "sf", "map")
) {
  ## read the meta data

  ## download the file, else use the package version
  meta <- import_isd_stations_live()
  return <- rlang::arg_match(return, c("table", "sf", "map"))

  # check year
  if (!any(end_year %in% c("current", "all"))) {
    if (!is.numeric(end_year)) {
      cli::cli_abort(
        "{.field end_year} should be one of 'current', 'all' or a numeric 4-digit year such as {2016}."
      )
    }
  }

  # we base the current year as the max available in the meta data
  if ("current" %in% end_year) {
    end_year <-
      max(as.numeric(format(meta$END, "%Y")), na.rm = TRUE)
  }
  if ("all" %in% end_year) {
    end_year <- 1900:2100
  }

  ## search based on name of site
  if (!is.null(site)) {
    ## search for station
    meta <- meta[grep(site, meta$STATION, ignore.case = TRUE), ]
  }

  ## search based on country codes
  if (!is.null(country)) {
    ## search for country
    id <- which(meta$CTRY %in% toupper(country))
    meta <- meta[id, ]
  }

  ## search based on state codes
  if (!is.null(state)) {
    ## search for state
    id <- which(meta$ST %in% toupper(state))
    meta <- meta[id, ]
  }

  # make sure no missing lat / lng
  id <- which(is.na(meta$LON))

  if (length(id) > 0) {
    meta <- meta[-id, ]
  }

  id <- which(is.na(meta$LAT))
  if (length(id) > 0) {
    meta <- meta[-id, ]
  }

  # filter by end year
  id <- which(format(meta$END, "%Y") %in% end_year)
  meta <- meta[id, ]

  ## approximate distance to site
  if (!is.null(lat) && !is.null(lng)) {
    point <-
      sf::st_as_sf(
        data.frame(lng = lng, lat = lat),
        coords = c("lng", "lat"),
        crs = sf::st_crs(crs)
      ) |>
      sf::st_transform(crs = sf::st_crs(4326))

    meta_sf <-
      sf::st_as_sf(meta, coords = c("LON", "LAT"), crs = 4326)

    meta$dist <- as.numeric(sf::st_distance(meta_sf, point)) / 1000L

    ## sort and return top n nearest
    meta <- dplyr::slice_min(meta, order_by = .data$dist, n = n_max)
  }

  dat <- dplyr::rename(meta, latitude = "LAT", longitude = "LON")

  names(dat) <- tolower(names(dat))

  meta <- dat

  if (return %in% c("sf", "map")) {
    meta <- sf::st_as_sf(
      meta,
      coords = c("longitude", "latitude"),
      crs = 4326,
      remove = FALSE
    )

    if (return == "map") {
      rlang::check_installed("leaflet")

      fmt_val <- function(x) {
        ifelse(
          is.na(x),
          "<span style='color: #bbb;'>N/A</span>",
          as.character(x)
        )
      }

      content <- paste0(
        "<div style='font-family: Arial, sans-serif; min-width: 220px; max-width: 280px;'>",

        # Combined name + code header
        "<div style='background: #2c9b6e; color: white; padding: 8px 12px; margin: -10px -10px 10px; border-radius: 4px 4px 0 0;'>",
        "<div style='font-size: 14px; font-weight: bold; margin-bottom: 4px;'>",
        dat$station,
        "</div>",
        "<div style='font-family: monospace; font-size: 12px; background: rgba(0,0,0,0.2); display: inline-block; padding: 2px 6px; border-radius: 3px; letter-spacing: 0.5px;'>",
        dat$code,
        "</div>",
        "</div>",

        # Temporal section
        "<div style='margin-bottom: 8px;'>",
        "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Coverage</div>",
        "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Start</span><span>",
        fmt_val(dat$begin),
        "</span></div>",
        "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>End</span><span>",
        fmt_val(dat$end),
        "</span></div>",
        "</div>",

        # Geography section
        "<div style='margin-bottom: 8px;'>",
        "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Geography</div>",
        "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Country</span><span>",
        fmt_val(dat$ctry),
        "</span></div>",
        "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>State</span><span>",
        fmt_val(dat$st),
        "</span></div>",
        "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Elevation</span><span>",
        ifelse(
          is.na(dat$`elev(m)`),
          fmt_val(dat$`elev(m)`),
          paste(dat$`elev(m)`, "m")
        ),
        "</span></div>",
        "</div>",
        "</div>"
      )

      if ("dist" %in% names(dat)) {
        content <- paste0(
          content,
          "<div style='margin-top: 8px;'>",
          "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Distance from search</div>",
          "<div style='font-family: monospace; font-size: 12px; background: #f4f4f4; border-left: 3px solid #c0392b; padding: 3px 8px; border-radius: 0 3px 3px 0;'>",
          round(dat$dist, 1),
          " km",
          "</div>",
          "</div>",
          "</div>" # closes the outer wrapper div
        )
      }

      m <- leaflet::leaflet(dat)

      if (!rlang::is_named(provider)) {
        provider <- stats::setNames(provider, provider)
      }
      for (i in seq_along(provider)) {
        m <- leaflet::addProviderTiles(
          m,
          provider = provider[[i]],
          group = names(provider)[[i]]
        )
      }

      m <-
        leaflet::addMarkers(
          map = m,
          ~longitude,
          ~latitude,
          popup = content,
          clusterOptions = leaflet::markerClusterOptions(),
          group = "Stations"
        )
      overlays <- "Stations"

      if (!is.null(lat) && !is.null(lng)) {
        overlays <- c(overlays, "Target")
        m <- leaflet::addAwesomeMarkers(
          map = m,
          data = point,
          group = "Target",
          icon = leaflet::makeAwesomeIcon(
            icon = "circle",
            library = "fa",
            markerColor = "red",
            iconColor = "#FFFFFF"
          ),
          popup = paste0(
            "<div style='font-family: Arial, sans-serif; min-width: 200px; max-width: 260px;'>",

            # Header
            "<div style='background: #c0392b; color: white; padding: 8px 12px; margin: -10px -10px 10px; border-radius: 4px 4px 0 0;'>",
            "<div style='font-size: 14px; font-weight: bold;'>Search Location</div>",
            "</div>",

            # Coordinates section
            "<div style='margin-bottom: 8px;'>",
            "<div style='font-size: 11px; font-weight: bold; text-transform: uppercase; color: #888; margin-bottom: 4px; border-bottom: 1px solid #eee; padding-bottom: 2px;'>Coordinates</div>",
            "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Latitude</span><span style='font-family: monospace;'>",
            lat,
            "</span></div>",
            "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>Longitude</span><span style='font-family: monospace;'>",
            lng,
            "</span></div>",
            "<div style='display: flex; justify-content: space-between; font-size: 12px; padding: 2px 0;'><span style='color: #555; font-weight: bold;'>CRS</span><span style='font-family: monospace;'>EPSG:",
            crs,
            "</span></div>",
            "</div>",

            "</div>"
          )
        )
      }

      m <-
        leaflet::addLayersControl(
          map = m,
          baseGroups = names(provider),
          overlayGroups = overlays,
          options = leaflet::layersControlOptions(
            collapsed = FALSE,
            autoZIndex = TRUE
          )
        )

      meta <- m
    }
  }

  return(meta)
}


#' Obtain site meta data from NOAA server
#'
#' Download all NOAA meta data, allowing for re-use and direct querying.
#'
#' @param ... Currently unused.
#'
#' @return a [tibble][tibble::tibble-package]
#'
#' @family NOAA ISD functions
#'
#' @examples
#' \dontrun{
#' meta <- import_isd_stations_live()
#' head(meta)
#' }
#' @export
import_isd_stations_live <- function(...) {
  ## downloads the whole thing fresh
  warn_isd()
  url <- "https://www.ncei.noaa.gov/pub/data/noaa/isd-history.csv"
  meta <- readr::read_csv(
    url,
    skip = 21,
    col_names = FALSE,
    col_types = readr::cols(
      X1 = readr::col_character(),
      X2 = readr::col_character(),
      X3 = readr::col_character(),
      X4 = readr::col_character(),
      X5 = readr::col_character(),
      X6 = readr::col_character(),
      X7 = readr::col_double(),
      X8 = readr::col_double(),
      X9 = readr::col_double(),
      X10 = readr::col_date(format = "%Y%m%d"),
      X11 = readr::col_date(format = "%Y%m%d")
    ),
    progress = FALSE
  )

  # if not available e.g. due to US Government shutdown, flag and exit
  # some header data may still be read, so check column number
  if (ncol(meta) == 1L) {
    cli::cli_abort(
      "File not available, check {.url https://www.ncei.noaa.gov/pub/data/noaa/} for potential server problems.",
      call. = FALSE
    )
  }

  ## names in the meta file
  names(meta) <- c(
    "USAF",
    "WBAN",
    "STATION",
    "CTRY",
    "ST",
    "CALL",
    "LAT",
    "LON",
    "ELEV(M)",
    "BEGIN",
    "END"
  )

  ## full character string of site id
  meta$USAF <-
    formatC(meta$USAF, width = 6, format = "d", flag = "0")

  ## code used to query data
  meta$code <- paste0(meta$USAF, "-", meta$WBAN)

  return(meta)
}

#' Import "Lite" Meteorological data from the NOAA Integrated Surface Database
#' (ISD)
#'
#' This function is an alternative to [importNOAA()], and provides access to the
#' "Lite" format of the data. This a subset of the larger [importNOAA()] dataset
#' featuring eight common climatological variables. As it assigns the nearest
#' measurement to the "top of the hour" to the data, specific values are likely
#' similar but different to those returned by [importNOAA()]. Read the
#' [technical
#' document](https://www.ncei.noaa.gov/pub/data/noaa/isd-lite/isd-lite-technical-document.pdf)
#' for more information.
#'
#' Note the following units for the main variables:
#'
#' \describe{
#'
#' \item{date}{Date/time in POSIXct format. **Note the time zone is UTC
#' and may need to be adjusted to merge with other local data.}
#'
#' \item{latitude}{Latitude in decimal degrees (-90 to 90).}
#'
#' \item{longitude}{Longitude in decimal degrees (-180 to 180). Negative numbers
#' are west of the Greenwich Meridian.}
#'
#' \item{elev}{Elevation of site in metres.}
#'
#' \item{ws}{Wind speed in m/s.}
#'
#' \item{wd}{Wind direction in degrees. 90 is from the east.}
#'
#' \item{air_temp}{Air temperature in degrees Celcius.}
#'
#' \item{atmos_pres}{The sea level pressure in millibars.}
#'
#' \item{dew_point}{The dew point temperature in degrees Celcius.}
#'
#' \item{precip_6}{6-hour precipitation in mm.}
#'
#' \item{precip_1}{1-hour precipitation in mm.}
#'
#' \item{sky}{Sky Condition Total Coverage Code.}
#' }
#'
#' The data are returned in GMT (UTC). It may be necessary to adjust the time
#' zone when combining with other data. For example, if air quality data were
#' available for Beijing with time zone set to "Etc/GMT-8" (note the negative
#' offset even though Beijing is ahead of GMT. See the `openair` package and
#' manual for more details), then the time zone of the met data can be changed
#' to be the same. One way of doing this would be `attr(met$date, "tzone") <-
#' "Etc/GMT-8"` for a meteorological data frame called `met`. The two data sets
#' could then be merged based on `date`.
#'
#' @inheritSection import_ghcn_hourly Parallel Processing
#'
#' @inheritParams import_isd_hourly
#' @export
#' @return Returns a data frame of surface observations. The data frame is
#'   consistent for use with the `openair` package. Note that the data are
#'   returned in GMT (UTC) time zone format. Users may wish to express the data
#'   in other time zones, e.g., to merge with air pollution data.
#'
#' @family NOAA ISD functions
#' @seealso [getMeta()] to obtain the codes based on various site search
#'   approaches.
#' @author Jack Davison
#' @examples
#'
#' \dontrun{
#' heathrow_lite <- import_isd_lite(code = "037720-99999", year = 2025)
#' }
import_isd_lite <- function(
  code = "037720-99999",
  year = 2025,
  progress = rlang::is_interactive(),
  quiet = FALSE
) {
  warn_isd()
  meta <- import_isd_stations(return = "table", end_year = "all")

  import_lite <- function(code, year) {
    station_name <- meta[meta$code == code, ]$station
    latitude <- meta[meta$code == code, ]$latitude
    longitude <- meta[meta$code == code, ]$longitude
    elevation <- meta[meta$code == code, ]$`elev(m)`

    path <- "https://www.ncei.noaa.gov/pub/data/noaa/isd-lite/DATAYEAR/DATACODE-DATAYEAR.gz"
    path <- gsub("DATACODE", code, path)
    path <- gsub("DATAYEAR", year, path)

    t <- tempfile(fileext = ".gz")

    download_success <- tryCatch(
      {
        suppressWarnings(utils::download.file(
          path,
          t,
          quiet = TRUE,
          mode = "wb"
        ))
        TRUE
      },
      error = function(e) {
        message(sprintf(
          "Download failed for code '%s' and year '%s': %s",
          code,
          year,
          e$message
        ))
        FALSE
      }
    )

    if (!download_success) {
      return(NULL)
    }

    utils::read.fwf(
      t,
      widths = c(5, 3, 3, 2, 6, 6, 6, 6, 6, 6, 6, 6),
      col.names = c(
        "year",
        "month",
        "day",
        "hour",
        "air_temp",
        "dew_point",
        "atmos_pres",
        "wd",
        "ws",
        "sky",
        "precip_1",
        "precip_6"
      )
    ) |>
      dplyr::tibble() |>
      dplyr::mutate(
        dplyr::across(
          dplyr::where(is.numeric),
          \(x) dplyr::na_if(x, -9999)
        )
      ) |>
      dplyr::mutate(
        date = ISOdate(
          year = .data$year,
          month = .data$month,
          day = .data$day,
          hour = .data$hour,
          tz = "UTC"
        ),
        .before = 0,
        .keep = "unused"
      ) |>
      dplyr::mutate(
        air_temp = .data$air_temp / 10,
        dew_point = .data$dew_point / 10,
        atmos_pres = .data$atmos_pres / 10,
        ws = .data$ws / 10,
        precip_1 = .data$precip_1 / 10,
        precip_6 = .data$precip_6 / 10
      ) |>
      dplyr::mutate(
        code = code,
        station = station_name,
        latitude = latitude,
        longitude = longitude,
        elev = elevation,
        .before = 0
      ) |>
      dplyr::relocate(
        "code",
        "station",
        "date",
        "latitude",
        "longitude",
        "elev",
        "ws",
        "wd",
        "air_temp",
        "atmos_pres",
        "dew_point",
        "precip_6",
        "precip_1",
        "sky"
      )
  }

  dat <-
    tidyr::crossing(
      code = code,
      year = year
    ) |>
    purrr::pmap(
      purrr::in_parallel(
        \(code, year) import_lite(code = code, year = year),
        import_lite = import_lite,
        meta = meta
      ),
      .progress = !quiet
    ) |>
    dplyr::bind_rows()

  if (is.null(dat) || nrow(dat) == 0) {
    cli::cli_inform(
      c(
        "x" = "Specified {.field site}-{.field year} combinations do not exist.",
        "i" = "Is the ISD service down? Check {.url https://www.ncei.noaa.gov/pub/data/noaa/isd-lite/}."
      )
    )
    return(NULL)
  }

  return(dat)
}

#' Export a meteorological data frame in ADMS format
#'
#' Writes a text file in the ADMS format to a location of the user's choosing,
#' with optional interpolation of missing values. This function works with data
#' from both [import_ghcn_hourly()] and [import_isd_hourly()].
#'
#' @param x A data frame imported by [import_ghcn_hourly()] or
#'   [import_isd_hourly()]. Note that this function only works for hourly GHCN
#'   data when `abbr_names = TRUE`.
#'
#' @param file A file name for the ADMS file. The file is written to the working
#'   directory by default.
#'
#' @param interp Should interpolation of missing values be undertaken? If `TRUE`
#'   linear interpolation is carried out for gaps of up to and including
#'   `max_gap`.
#'
#' @param max_gap The maximum gap in hours that should be interpolated where
#'   there are missing data when `interp = TRUE.` Data with gaps more than
#'   `max_gap` are left as missing.
#'
#' @return `write_adms()` returns the input `dat` invisibly.
#' @family Met writing functions
#' @family ADMS functions
#' @export
#' @examples
#' \dontrun{
#' # import some data then export it
#' dat <- import_isd_hourly(year = 2012)
#' write_adms(dat, file = "~/adms_met.MET")
#' }
write_adms <- function(
  x,
  file = "./ADMS_met.MET",
  interp = FALSE,
  max_gap = 2
) {
  # save input for later
  input <- x

  # ensure all data is lowercase - difference between ISD & GHCN
  dat <- dplyr::rename_with(x, .fn = tolower)

  # make sure the data do not have gaps
  all_dates <- dplyr::tibble(
    date = seq(
      ISOdatetime(
        year = as.numeric(format(dat$date[1], "%Y")),
        month = 1,
        day = 1,
        hour = 0,
        min = 0,
        sec = 0,
        tz = "GMT"
      ),
      ISOdatetime(
        year = as.numeric(format(dat$date[1], "%Y")),
        month = 12,
        day = 31,
        hour = 23,
        min = 0,
        sec = 0,
        tz = "GMT"
      ),
      by = "hour"
    )
  )

  # complete dataset
  dat <- dplyr::left_join(all_dates, dat, by = dplyr::join_by("date"))

  # interpolate if requested
  if (interp) {
    var_interp <- c("ws", "u", "v", "air_temp", "rh", "cl")

    # transform wd
    dat <- dplyr::mutate(
      dat,
      u = sin(pi * .data$wd / 180),
      v = cos(pi * .data$wd / 180)
    )

    for (variable in var_interp) {
      # if all missing, then don't interpolate
      if (all(is.na(dat[[variable]]))) {
        return()
      }

      # first fill with linear interpolation
      filled <- stats::approx(
        dat$date,
        dat[[variable]],
        xout = dat$date,
        na.rm = TRUE,
        rule = 2,
        method = "linear"
      )$y

      # find out length of missing data
      is_missing <- rle(is.na(dat[[variable]]))

      is_missing <- rep(
        ifelse(is_missing$values, is_missing$lengths, 0),
        times = is_missing$lengths
      )

      id <- which(is_missing > max_gap)

      # update data frame
      dat[[variable]] <- filled
      dat[[variable]][id] <- NA
    }

    dat <- dplyr::mutate(
      dat,
      wd = as.vector(atan2(.data$u, .data$v) * 360 / 2 / pi)
    )

    # correct for negative wind directions
    ids <- which(dat$wd < 0) ## ids where wd < 0
    dat$wd[ids] <- dat$wd[ids] + 360

    dat <- dplyr::select(dat, -"v", -"u")
  }

  # check if present
  if (!"cl" %in% names(dat)) {
    dat$cl <- NA
  }
  if (!"precip" %in% names(dat)) {
    dat$precip <- NA
  }

  # reformat for ADMS output
  adms <-
    dat |>
    # select needed variables
    dplyr::transmute(
      station = "0000",
      year = as.numeric(format(.data$date, "%Y")),
      day = as.numeric(format(.data$date, "%j")),
      hour = as.numeric(format(.data$date, "%H")),
      air_temp = round(.data$air_temp, 1),
      ws = round(.data$ws, 1),
      wd = round(.data$wd, 1),
      rh = round(.data$rh, 1),
      cl = round(.data$cl, 0),
      precip = round(.data$precip, 1)
    ) |>
    # replace NA with -999
    dplyr::mutate(
      dplyr::across(
        dplyr::where(is.numeric),
        \(x) tidyr::replace_na(x, -999)
      )
    )

  # message key data capture rates
  calc_dc <- function(x) {
    round(100 * mean(x != -999), 1)
  }
  cli::cli_inform(
    c(
      "i" = "Data capture for {.strong wind speed}: {calc_dc(adms$ws)}%",
      "i" = "Data capture for {.strong wind direction}: {calc_dc(adms$wd)}%",
      "i" = "Data capture for {.strong temperature}: {calc_dc(adms$air_temp)}%",
      "i" = "Data capture for {.strong cloud cover}: {calc_dc(adms$cl)}%"
    )
  )

  # write the data file
  utils::write.table(
    adms,
    file = file,
    col.names = FALSE,
    row.names = FALSE,
    sep = ",",
    quote = FALSE
  )

  # add the header lines
  fConn <- file(file, "r+")
  Lines <- readLines(fConn)
  writeLines(
    c(
      "VARIABLES:\n10\nSTATION DCNN\nYEAR\nTDAY\nTHOUR\nT0C\nU\nPHI\nRHUM\nCL\nP\nDATA:",
      Lines
    ),
    con = fConn
  )
  close(fConn)

  # return input invisibly
  invisible(input)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_ghcn_stations.R
\name{import_ghcn_stations}
\alias{import_ghcn_stations}
\title{Import station metadata for the Global Historical Climatology Network}
\usage{
import_ghcn_stations(
  name = NULL,
  country = NULL,
  state = NULL,
  lat = NULL,
  lng = NULL,
  crs = 4326,
  n_max = 10L,
  provider = c(`Street Map` = "CartoDB.Voyager", Satellite = "Esri.WorldImagery"),
  database = c("hourly", "daily"),
  return = c("table", "sf", "map")
)
}
\arguments{
\item{name, country, state}{String values to use to filter the metadata for
specific site names, countries and states. \code{country} and \code{state} are
matched exactly to codes accessed using \code{\link[=import_ghcn_countries]{import_ghcn_countries()}}. \code{name}
is searched as a sub-string case insensitively.}

\item{lat, lng, n_max}{Decimal latitude (\code{lat}) and longitude (\code{lng}) (or other
Y/X coordinate if using a different \code{crs}). If provided, the \code{n_max}
closest ISD stations to this coordinate will be returned.}

\item{crs}{The coordinate reference system (CRS) of the data, passed to
\code{\link[sf:st_crs]{sf::st_crs()}}. By default this is \href{https://epsg.io/4326}{EPSG:4326}, the
CRS associated with the commonly used latitude and longitude coordinates.
Different coordinate systems can be specified using \code{crs} (e.g., \code{crs = 27700} for the \href{https://epsg.io/27700}{British National Grid}). Note that
non-lat/lng coordinate systems will be re-projected to \code{EPSG:4326} for
making comparisons with the NOAA metadata.}

\item{provider}{When \code{return = "map"}, by default a map will be created in
which readers may toggle between a vector street map and a satellite/aerial
image. \code{provider} allows users to override this default; see
\url{http://leaflet-extras.github.io/leaflet-providers/preview/} for a list
of all base maps that can be used. Base maps can be toggled using a layer
control menu; the labels will be taken from the name of the base map unless
a named list is defined (see default value).}

\item{database}{One of \code{"hourly"} or \code{"daily"}, which defines whether to
import stations available in the GHCNh or GHCNd. Note that there is overlap
between the two, but some stations may only be available in one or the
other.}

\item{return}{The type of R object to import the data as. One of the
following:
\itemize{
\item \code{"table"}, which returns an R \code{data.frame}.
\item \code{"sf"}, which returns a spatial \code{data.frame} from the \code{sf} package.
\item \code{"map"}, which returns an interactive \code{leaflet} map.
}}
}
\value{
One of:
\itemize{
\item a \link[tibble:tibble-package]{tibble}
\item an \link[sf:st_as_sf]{sf} object
\item an interactive \code{leaflet} map
}
}
\description{
This function accesses a full list of GHCN stations available through either
the GHCNh or GHCNd. As well as the station \code{id}, needed for importing
measurement data, useful geographic and network metadata is also returned.
}
\seealso{
Other GHCN functions:
\code{\link[=import_ghcn_countries]{import_ghcn_countries()}},
\code{\link[=import_ghcn_daily]{import_ghcn_daily()}},
\code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}},
\code{\link[=import_ghcn_inventory]{import_ghcn_inventory()}},
\code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}}
}
\author{
Jack Davison
}
\concept{GHCN functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weatherCodes.R
\docType{data}
\name{weatherCodes}
\alias{weatherCodes}
\title{Codes for weather types}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 100 rows and 2 columns.
}
\usage{
weatherCodes
}
\description{
This data frame consists of the weather description codes used in the ISD. It
is not of general use to most users.
}
\details{
\describe{
\item{pwc}{Weather code, which can be merged with the \code{pwc} column in \code{\link[=importNOAA]{importNOAA()}} datasets.}
\item{description}{Description associated with the weather codes.}
}
}
\examples{
weatherCodes
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_isd_lite.R
\name{import_isd_lite}
\alias{import_isd_lite}
\title{Import "Lite" Meteorological data from the NOAA Integrated Surface Database
(ISD)}
\usage{
import_isd_lite(
  code = "037720-99999",
  year = 2025,
  progress = rlang::is_interactive(),
  quiet = FALSE
)
}
\arguments{
\item{code}{The identifying code as a character string. The code is a
combination of the USAF and the WBAN unique identifiers. The codes are
separated by a \dQuote{-} e.g. \code{code = "037720-99999"}.}

\item{year}{The year to import. This can be a vector of years e.g. \code{year = 2000:2005}.}

\item{progress}{Show a progress bar when importing many stations/years?
Defaults to \code{TRUE} in interactive R sessions. Passed to \code{.progress} in
\code{\link[purrr:map]{purrr::map()}} and/or \code{\link[purrr:pmap]{purrr::pmap()}}.}

\item{quiet}{Print missing sites/years to the screen? Defaults to \code{FALSE}.}
}
\value{
Returns a data frame of surface observations. The data frame is
consistent for use with the \code{openair} package. Note that the data are
returned in GMT (UTC) time zone format. Users may wish to express the data
in other time zones, e.g., to merge with air pollution data.
}
\description{
This function is an alternative to \code{\link[=importNOAA]{importNOAA()}}, and provides access to the
"Lite" format of the data. This a subset of the larger \code{\link[=importNOAA]{importNOAA()}} dataset
featuring eight common climatological variables. As it assigns the nearest
measurement to the "top of the hour" to the data, specific values are likely
similar but different to those returned by \code{\link[=importNOAA]{importNOAA()}}. Read the
\href{https://www.ncei.noaa.gov/pub/data/noaa/isd-lite/isd-lite-technical-document.pdf}{technical document}
for more information.
}
\details{
Note the following units for the main variables:

\describe{

\item{date}{Date/time in POSIXct format. **Note the time zone is UTC
and may need to be adjusted to merge with other local data.}

\item{latitude}{Latitude in decimal degrees (-90 to 90).}

\item{longitude}{Longitude in decimal degrees (-180 to 180). Negative numbers
are west of the Greenwich Meridian.}

\item{elev}{Elevation of site in metres.}

\item{ws}{Wind speed in m/s.}

\item{wd}{Wind direction in degrees. 90 is from the east.}

\item{air_temp}{Air temperature in degrees Celcius.}

\item{atmos_pres}{The sea level pressure in millibars.}

\item{dew_point}{The dew point temperature in degrees Celcius.}

\item{precip_6}{6-hour precipitation in mm.}

\item{precip_1}{1-hour precipitation in mm.}

\item{sky}{Sky Condition Total Coverage Code.}
}

The data are returned in GMT (UTC). It may be necessary to adjust the time
zone when combining with other data. For example, if air quality data were
available for Beijing with time zone set to "Etc/GMT-8" (note the negative
offset even though Beijing is ahead of GMT. See the \code{openair} package and
manual for more details), then the time zone of the met data can be changed
to be the same. One way of doing this would be \code{attr(met$date, "tzone") <- "Etc/GMT-8"} for a meteorological data frame called \code{met}. The two data sets
could then be merged based on \code{date}.
}
\section{Parallel Processing}{


If you are importing a lot of meteorological data, this can take a long
while. This is because each combination of year and station requires
downloading a separate data file from NOAA's online data directory, and the
time each download takes can quickly add up. Many data import functions in
\code{{worldmet}} can use parallel processing to speed downloading up, powered
by the capable \code{{mirai}} package. If users have any \code{{mirai}} "daemons"
set, these functions will download files in parallel. The greatest benefits
will be seen if you spawn as many daemons as you have cores on your
machine, although one fewer than the available cores is often a good rule
of thumb. Your mileage may vary, however, and naturally spawning more
daemons than station-year combinations will lead to diminishing returns.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set workers - once per session
mirai::daemons(4)

# import lots of data - NB: no change in the import function!
big_met <- import_ghcn_hourly(code = "UKI0000EGLL", year = 2010:2025)
}\if{html}{\out{</div>}}
}

\examples{

\dontrun{
heathrow_lite <- import_isd_lite(code = "037720-99999", year = 2025)
}
}
\seealso{
\code{\link[=getMeta]{getMeta()}} to obtain the codes based on various site search
approaches.

Other NOAA ISD functions:
\code{\link[=import_isd_hourly]{import_isd_hourly()}},
\code{\link[=import_isd_stations]{import_isd_stations()}},
\code{\link[=import_isd_stations_live]{import_isd_stations_live()}}
}
\author{
Jack Davison
}
\concept{NOAA ISD functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_isd_stations.R
\name{import_isd_stations}
\alias{import_isd_stations}
\title{Import station metadata for the Integrated Surface Databse}
\usage{
import_isd_stations(
  site = NULL,
  country = NULL,
  state = NULL,
  lat = NULL,
  lng = NULL,
  crs = 4326,
  n_max = 10,
  end_year = "current",
  provider = c(`Street Map` = "CartoDB.Voyager", Satellite = "Esri.WorldImagery"),
  return = c("table", "sf", "map")
)
}
\arguments{
\item{site}{A site name search string e.g. \code{site = "heathrow"}. The search
strings and be partial and can be upper or lower case e.g. \code{site = "HEATHR"}.}

\item{country}{The country code. This is a two letter code. For a full
listing see \url{https://www.ncei.noaa.gov/pub/data/noaa/isd-history.csv}.}

\item{state}{The state code. This is a two letter code.}

\item{lat, lng, n_max}{Decimal latitude (\code{lat}) and longitude (\code{lng}) (or other
Y/X coordinate if using a different \code{crs}). If provided, the \code{n_max}
closest ISD stations to this coordinate will be returned.}

\item{crs}{The coordinate reference system (CRS) of the data, passed to
\code{\link[sf:st_crs]{sf::st_crs()}}. By default this is \href{https://epsg.io/4326}{EPSG:4326}, the
CRS associated with the commonly used latitude and longitude coordinates.
Different coordinate systems can be specified using \code{crs} (e.g., \code{crs = 27700} for the \href{https://epsg.io/27700}{British National Grid}). Note that
non-lat/lng coordinate systems will be re-projected to \code{EPSG:4326} for
making comparisons with the NOAA metadata.}

\item{end_year}{To help filter sites based on how recent the available data
are. \code{end_year} can be "current", "any" or a numeric year such as 2016, or
a range of years e.g. 1990:2016 (which would select any site that had an
end date in that range. \strong{By default only sites that have some data for the
current year are returned}.}

\item{provider}{When \code{return = "map"}, by default a map will be created in
which readers may toggle between a vector street map and a satellite/aerial
image. \code{provider} allows users to override this default; see
\url{http://leaflet-extras.github.io/leaflet-providers/preview/} for a list
of all base maps that can be used. Base maps can be toggled using a layer
control menu; the labels will be taken from the name of the base map unless
a named list is defined (see default value).}

\item{return}{The type of R object to import the data as. One of the
following:
\itemize{
\item \code{"table"}, which returns an R \code{data.frame}.
\item \code{"sf"}, which returns a spatial \code{data.frame} from the \code{sf} package.
\item \code{"map"}, which returns an interactive \code{leaflet} map.
}}
}
\value{
A data frame is returned with all available meta data, mostly
importantly including a \code{code} that can be supplied to \code{\link[=importNOAA]{importNOAA()}}. If
latitude and longitude searches are made an approximate distance, \code{dist} in
km is also returned.
}
\description{
This function is primarily used to find a site code that can be used to
access data using \code{\link[=import_isd_hourly]{import_isd_hourly()}}. Sites searches of approximately
30,000 sites can be carried out based on the site name and based on the
nearest locations based on user-supplied latitude and longitude.
}
\examples{
\dontrun{
## search for sites with name beijing
getMeta(site = "beijing")
}

\dontrun{
## search for near a specified lat/lng - near Beijing airport
## returns 'n_max' nearest by default
getMeta(lat = 40, lng = 116.9)
}
}
\seealso{
Other NOAA ISD functions:
\code{\link[=import_isd_hourly]{import_isd_hourly()}},
\code{\link[=import_isd_lite]{import_isd_lite()}},
\code{\link[=import_isd_stations_live]{import_isd_stations_live()}}
}
\author{
David Carslaw
}
\concept{NOAA ISD functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_ghcn_stations.R
\name{import_ghcn_inventory}
\alias{import_ghcn_inventory}
\title{Import station inventory for the Global Historical Climatology Network}
\usage{
import_ghcn_inventory(
  database = c("hourly", "daily"),
  pivot = c("wide", "long"),
  progress = rlang::is_interactive()
)
}
\arguments{
\item{database}{One of \code{"hourly"} or \code{"daily"}, which defines whether to
import the GHCNh or GHCNd inventory. The way in which these files is
formatted is different.}

\item{pivot}{One of \code{"wide"} or \code{"long"}. The GHCNh inventory can be returned
in a \code{"wide"} format (with \code{id}, \code{year} and twelve month columns) or a
\code{"long"} format (with \code{id}, \code{year}, \code{month}, and \code{count} columns). Does not
apply to the GHCNd inventory.}

\item{progress}{The inventory file is large and can be slow to download. Show
a progress indicator when accessing the inventory? Defaults to \code{TRUE} in
interactive R sessions. Passed to \code{progress} in \code{\link[readr:read_fwf]{readr::read_fwf()}} and/or
\code{\link[purrr:pmap]{purrr::pmap()}}.}
}
\value{
a \link[tibble:tibble-package]{tibble}
}
\description{
This function accesses a data inventory of GHCN stations available through
either the GHCNh or GHCNd. The returned \code{data.frame} contains data which
reveals the earliest and latest years of data available for each station from
the NOAA database.
}
\seealso{
Other GHCN functions:
\code{\link[=import_ghcn_countries]{import_ghcn_countries()}},
\code{\link[=import_ghcn_daily]{import_ghcn_daily()}},
\code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}},
\code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}},
\code{\link[=import_ghcn_stations]{import_ghcn_stations()}}
}
\author{
Jack Davison
}
\concept{GHCN functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_ghcn_stations.R
\name{import_ghcn_countries}
\alias{import_ghcn_countries}
\title{Import FIPS country codes and State/Province/Territory codes used by the
Global Historical Climatology Network}
\usage{
import_ghcn_countries(
  table = c("countries", "states"),
  database = c("hourly", "daily", "monthly")
)
}
\arguments{
\item{table}{One of \code{"countries"} or \code{"states"}.}

\item{database}{One of \code{"hourly"}, \code{"daily"} or \code{"monthly"}, which defines
which of the NOAA databases to import the FIPS codes from. There is little
difference between the data in the different sources, but this option may
be useful if one of the services is not accessible.}
}
\value{
a \link[tibble:tibble-package]{tibble}
}
\description{
This function returns a two-column dataframe either of "Federal Information
Processing Standards" (FIPS) codes and the countries to which they are
associated, or state codes and their associated states. These may be a useful
reference when examining GHCN site metadata.
}
\seealso{
Other GHCN functions:
\code{\link[=import_ghcn_daily]{import_ghcn_daily()}},
\code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}},
\code{\link[=import_ghcn_inventory]{import_ghcn_inventory()}},
\code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}},
\code{\link[=import_ghcn_stations]{import_ghcn_stations()}}
}
\author{
Jack Davison
}
\concept{GHCN functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_ghcn_daily.R
\name{import_ghcn_daily}
\alias{import_ghcn_daily}
\title{Import data from the Global Historical Climatology daily (GHCNd) database}
\usage{
import_ghcn_daily(
  station,
  year = NULL,
  source = c("csv"),
  extra = FALSE,
  append_codes = FALSE,
  codes = c("measurement_flag", "quality_flag", "source_flag"),
  progress = rlang::is_interactive()
)
}
\arguments{
\item{station}{One or more site codes for the station(s) of interest,
obtained using \code{\link[=import_ghcn_stations]{import_ghcn_stations()}}.}

\item{year}{One or more years of interest. If \code{NULL}, the default, all years
of data available for the chosen \code{station}s will be imported. Note that, in
the GHCNd and GHCNm, files are split by station but not year, so setting a
\code{year} will not speed up the download. Specifying fewer years will improve
the speed of a GHCNh download, however.}

\item{source}{The data format for the GHCNd. Currently only \code{"csv"} is
supported. This argument is included for future use.}

\item{extra}{Should additional columns be returned? The default, \code{FALSE},
returns an opinionated selection of elements that'll be of most interest to
most users. \code{TRUE} will return everything available.}

\item{append_codes}{Logical. Should various codes and flags be appended to
the output dataframe?}

\item{codes}{When \code{append_codes} is \code{TRUE}, which codes should be appended to
the dataframe? Any combination of \code{"measurement_flag"}, \code{"quality_flag"},
and/or \code{"source_flag"}.}

\item{progress}{Show a progress bar when importing many stations/years?
Defaults to \code{TRUE} in interactive R sessions. Passed to \code{.progress} in
\code{\link[purrr:map]{purrr::map()}} and/or \code{\link[purrr:pmap]{purrr::pmap()}}.}
}
\value{
a \link[tibble:tibble-package]{tibble}
}
\description{
This function flexibly accesses meteorological data from the GHCNd database.
Users can provide any of stations, and control whether attribute codes are
returned with the data.
}
\section{Data Definition}{


The core elements in the GHCNd are:
\itemize{
\item \strong{PRCP}: Precipitation (mm)
\item \strong{SNOW}: Snowfall (mm)
\item \strong{SNWD}: Snow depth (mm)
\item \strong{TMAX}: Maximum temperature (degrees C)
\item \strong{TMIN}: Minimum temperature (degrees C)
}

Other elements which may appear are:
\itemize{
\item \strong{ACMC}: Average cloudiness midnight to midnight from 30-second ceilometer data (percent)
\item \strong{ACMH}: Average cloudiness midnight to midnight from manual observations (percent)
\item \strong{ACSC}: Average cloudiness sunrise to sunset from 30-second ceilometer data (percent)
\item \strong{ACSH}: Average cloudiness sunrise to sunset from manual observations (percent)
\item \strong{ADPT}: Average Dew Point Temperature for the day (degrees C)
\item \strong{ASLP}: Average Sea Level Pressure for the day (hPa)
\item \strong{ASTP}: Average Station Level Pressure for the day (hPa)
\item \strong{AWBT}: Average Wet Bulb Temperature for the day (degrees C)
\item \strong{AWDR}: Average daily wind direction (degrees)
\item \strong{AWND}: Average daily wind speed (meters per second)
\item \strong{DAEV}: Number of days included in the multiday evaporation total (MDEV)
\item \strong{DAPR}: Number of days included in the multiday precipitation total (MDPR)
\item \strong{DASF}: Number of days included in the multiday snowfall total (MDSF)
\item \strong{DATN}: Number of days included in the multiday minimum temperature (MDTN)
\item \strong{DATX}: Number of days included in the multiday maximum temperature (MDTX)
\item \strong{DAWM}: Number of days included in the multiday wind movement (MDWM)
\item \strong{DWPR}: Number of days with non-zero precipitation included in multiday precipitation total (MDPR)
\item \strong{EVAP}: Evaporation of water from evaporation pan (mm)
\item \strong{FMTM}: Time of fastest mile or fastest 1-minute wind (hours and minutes, i.e., HHMM)
\item \strong{FRGB}: Base of frozen ground layer (cm)
\item \strong{FRGT}: Top of frozen ground layer (cm)
\item \strong{FRTH}: Thickness of frozen ground layer (cm)
\item \strong{GAHT}: Difference between river and gauge height (cm)
\item \strong{MDEV}: Multiday evaporation total (mm; use with \strong{DAEV})
\item \strong{MDPR}: Multiday precipitation total (mm; use with \strong{DAPR} and \strong{DWPR}, if available)
\item \strong{MDSF}: Multiday snowfall total
\item \strong{MDTN}: Multiday minimum temperature (degrees C; use with \strong{DATN})
\item \strong{MDTX}: Multiday maximum temperature (degrees C; use with \strong{DATX})
\item \strong{MDWM}: Multiday wind movement (km)
\item \strong{MNPN}: Daily minimum temperature of water in an evaporation pan (degrees C)
\item \strong{MXPN}: Daily maximum temperature of water in an evaporation pan (degrees C)
\item \strong{PGTM}: Peak gust time (hours and minutes, i.e., HHMM)
\item \strong{PSUN}: Daily percent of possible sunshine (percent)
\item \strong{RHAV}: Average relative humidity for the day (percent)
\item \strong{RHMN}: Minimum relative humidity for the day (percent)
\item \strong{RHMX}: Maximum relative humidity for the day (percent)
\item \strong{SN$#}: Minimum soil temperature (degrees C) where $ corresponds to a code for ground cover and # corresponds to a code for soil depth. Ground cover codes: 0=unknown, 1=grass, 2=fallow, 3=bare ground, 4=brome grass, 5=sod, 6=straw mulch, 7=grass muck, 8=bare muck. Depth codes: 1=5cm, 2=10cm, 3=20cm, 4=50cm, 5=100cm, 6=150cm, 7=180cm
\item \strong{SX$#}: Maximum soil temperature (degrees C) where $ corresponds to a code for ground cover and # corresponds to a code for soil depth. See \strong{SN$#} for ground cover and depth codes
\item \strong{TAXN}: Average daily temperature computed as \code{(TMAX+TMIN)/2.0} (degrees C)
\item \strong{TAVG}: Average daily temperature (degrees C). Note that TAVG from source 'S' corresponds to an average of hourly readings for the period ending at 2400 UTC rather than local midnight or other Local Standard Time according to a specific Met Service's protocol. For sources other than 'S' TAVG is computed in a variety of ways including traditional fixed hours of the day whereas TAXN is solely computed as (TMAX+TMIN)/2.0
\item \strong{THIC}: Thickness of ice on water (mm)
\item \strong{TOBS}: Temperature at the time of observation (degrees C)
\item \strong{TSUN}: Daily total sunshine (minutes)
\item \strong{WDF1}: Direction of fastest 1-minute wind (degrees)
\item \strong{WDF2}: Direction of fastest 2-minute wind (degrees)
\item \strong{WDF5}: Direction of fastest 5-second wind (degrees)
\item \strong{WDFG}: Direction of peak wind gust (degrees)
\item \strong{WDFI}: Direction of highest instantaneous wind (degrees)
\item \strong{WDFM}: Fastest mile wind direction (degrees)
\item \strong{WDMV}: 24-hour wind movement (km)
\item \strong{WESD}: Water equivalent of snow on the ground (mm)
\item \strong{WESF}: Water equivalent of snowfall (mm)
\item \strong{WSF1}: Fastest 1-minute wind speed (meters per second)
\item \strong{WSF2}: Fastest 2-minute wind speed (meters per second)
\item \strong{WSF5}: Fastest 5-second wind speed (meters per second)
\item \strong{WSFG}: Peak gust wind speed (meters per second)
\item \strong{WSFI}: Highest instantaneous wind speed (meters per second)
\item \strong{WSFM}: Fastest mile wind speed (meters per second)
}

There can be any number of weather types (\strong{WT$$})
\itemize{
\item \strong{WT01}: Fog, ice fog, or freezing fog (may include heavy fog)
\item \strong{WT02}: Heavy fog or heaving freezing fog (not always distinguished from fog)
\item \strong{WT03}: Thunder
\item \strong{WT04}: Ice pellets, sleet, snow pellets, or small hail
\item \strong{WT05}: Hail (may include small hail)
\item \strong{WT06}: Glaze or rime
\item \strong{WT07}: Dust, volcanic ash, blowing dust, blowing sand, or blowing obstruction
\item \strong{WT08}: Smoke or haze
\item \strong{WT09}: Blowing or drifting snow
\item \strong{WT10}: Tornado, waterspout, or funnel cloud
\item \strong{WT11}: High or damaging winds
\item \strong{WT12}: Blowing spray
\item \strong{WT13}: Mist
\item \strong{WT14}: Drizzle
\item \strong{WT15}: Freezing drizzle
\item \strong{WT16}: Rain (may include freezing rain, drizzle, and freezing drizzle)
\item \strong{WT17}: Freezing rain
\item \strong{WT18}: Snow, snow pellets, snow grains, or ice crystals
\item \strong{WT19}: Unknown source of precipitation
\item \strong{WT21}: Ground fog
\item \strong{WT22}: Ice fog or freezing fog
}

There can also be any number of 'weather in the vicinity' columns
(\strong{WV$$})
\itemize{
\item \strong{WV01}: Fog, ice fog, or freezing fog (may include heavy fog)
\item \strong{WV03}: Thunder
\item \strong{WV07}: Ash, dust, sand, or other blowing obstruction
\item \strong{WV18}: Snow or ice crystals
\item \strong{WV20}: Rain or snow shower
}
}

\section{Parallel Processing}{


If you are importing a lot of meteorological data, this can take a long
while. This is because each combination of year and station requires
downloading a separate data file from NOAA's online data directory, and the
time each download takes can quickly add up. Many data import functions in
\code{{worldmet}} can use parallel processing to speed downloading up, powered
by the capable \code{{mirai}} package. If users have any \code{{mirai}} "daemons"
set, these functions will download files in parallel. The greatest benefits
will be seen if you spawn as many daemons as you have cores on your
machine, although one fewer than the available cores is often a good rule
of thumb. Your mileage may vary, however, and naturally spawning more
daemons than station-year combinations will lead to diminishing returns.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set workers - once per session
mirai::daemons(4)

# import lots of data - NB: no change in the import function!
big_met <- import_ghcn_hourly(code = "UKI0000EGLL", year = 2010:2025)
}\if{html}{\out{</div>}}
}

\seealso{
Other GHCN functions:
\code{\link[=import_ghcn_countries]{import_ghcn_countries()}},
\code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}},
\code{\link[=import_ghcn_inventory]{import_ghcn_inventory()}},
\code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}},
\code{\link[=import_ghcn_stations]{import_ghcn_stations()}}
}
\author{
Jack Davison
}
\concept{GHCN functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_met.R
\name{write_met}
\alias{write_met}
\title{Export a meteorological data frame in files, chunked by site and year}
\usage{
write_met(
  x,
  path = ".",
  ext = c("rds", "delim", "parquet"),
  delim = ",",
  suffix = "",
  progress = rlang::is_interactive()
)
}
\arguments{
\item{x}{A data frame imported by \code{\link[=import_isd_hourly]{import_isd_hourly()}},
\code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}}, or \code{\link[=import_ghcn_daily]{import_ghcn_daily()}}.}

\item{path}{The path to a directory to save each file. By default, this is
the working directory.}

\item{ext}{The file type to use when saving the data. Can be \code{"rds"},
\code{"delim"} or \code{"parquet"}. Note that \code{"parquet"} requires the \code{arrow}
package.}

\item{delim}{Delimiter used to separate values when \code{ext = "delim"}. Must be
a single character. Defaults to being comma-delimited (\code{","}).}

\item{suffix}{An additional suffix to append to file names. Useful examples
could be \code{"_isd"}, \code{"_hourly"}, \code{"_lite"}, and so on.}

\item{progress}{Show a progress bar when writing many stations/years?
Defaults to \code{TRUE} in interactive R sessions. Passed to \code{.progress} in
\code{\link[purrr:walk]{purrr::walk()}}.}
}
\value{
\code{write_met()} returns \code{path} invisibly.
}
\description{
Writes data returned by any of \code{\link[=import_isd_hourly]{import_isd_hourly()}}, \code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}},
or \code{\link[=import_ghcn_daily]{import_ghcn_daily()}} to a file. Each station and year in the data is
written to a separate file.
}
\examples{
\dontrun{
# import some data then export it
dat <- import_isd_hourly(year = 2012)
write_met(dat)
}
}
\seealso{
Other Met writing functions:
\code{\link[=write_adms]{write_adms()}}
}
\concept{Met writing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/worldmet-package.R
\docType{package}
\name{worldmet-package}
\alias{worldmet}
\alias{worldmet-package}
\title{worldmet: Import Surface Meteorological Data from NOAA}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Functions to import data from more than 30,000 surface meteorological sites around the world managed by the National Oceanic and Atmospheric Administration (NOAA) Global Historical Climate Network (GHCN) and Integrated Surface Database (ISD).
}
\details{
This package contains functions to import surface meteorological
data from over 30,000 sites around the world. These data are curated by
NOAA as part of the Global Historical Climate Network (GHCN), which
replaced the Integrated Surface Database (ISD) in 2025.

If you access these data using the \code{worldmet} package please give full
acknowledgement to NOAA. Users should also take a note of the usage
restrictions.

These data work well with the \code{openair} package that has been developed to
analyse air pollution data.
}
\references{
For general information about the GHCNh, see
\url{https://www.ncei.noaa.gov/products/global-historical-climatology-network-hourly}.

For general information about the legacy ISD service, see
\url{https://www.ncei.noaa.gov/products/land-based-station/integrated-surface-database}.
}
\seealso{
See \url{https://github.com/openair-project/openair} for information on
the related \code{openair} package.
}
\author{
\strong{Maintainer}: David Carslaw \email{david.carslaw@york.ac.uk} (\href{https://orcid.org/0000-0003-0991-950X}{ORCID})

Authors:
\itemize{
  \item David Carslaw \email{david.carslaw@york.ac.uk} (\href{https://orcid.org/0000-0003-0991-950X}{ORCID})
  \item Jack Davison \email{jack.davison@ricardo.com} (\href{https://orcid.org/0000-0003-2653-6615}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_isd_stations.R
\name{import_isd_stations_live}
\alias{import_isd_stations_live}
\title{Obtain site meta data from NOAA server}
\usage{
import_isd_stations_live(...)
}
\arguments{
\item{...}{Currently unused.}
}
\value{
a \link[tibble:tibble-package]{tibble}
}
\description{
Download all NOAA meta data, allowing for re-use and direct querying.
}
\examples{
\dontrun{
meta <- import_isd_stations_live()
head(meta)
}
}
\seealso{
Other NOAA ISD functions:
\code{\link[=import_isd_hourly]{import_isd_hourly()}},
\code{\link[=import_isd_lite]{import_isd_lite()}},
\code{\link[=import_isd_stations]{import_isd_stations()}}
}
\concept{NOAA ISD functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_adms.R
\name{write_adms}
\alias{write_adms}
\title{Export a meteorological data frame in ADMS format}
\usage{
write_adms(x, file = "./ADMS_met.MET", interp = FALSE, max_gap = 2)
}
\arguments{
\item{x}{A data frame imported by \code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}} or
\code{\link[=import_isd_hourly]{import_isd_hourly()}}. Note that this function only works for hourly GHCN
data when \code{abbr_names = TRUE}.}

\item{file}{A file name for the ADMS file. The file is written to the working
directory by default.}

\item{interp}{Should interpolation of missing values be undertaken? If \code{TRUE}
linear interpolation is carried out for gaps of up to and including
\code{max_gap}.}

\item{max_gap}{The maximum gap in hours that should be interpolated where
there are missing data when \code{interp = TRUE.} Data with gaps more than
\code{max_gap} are left as missing.}
}
\value{
\code{write_adms()} returns the input \code{dat} invisibly.
}
\description{
Writes a text file in the ADMS format to a location of the user's choosing,
with optional interpolation of missing values. This function works with data
from both \code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}} and \code{\link[=import_isd_hourly]{import_isd_hourly()}}.
}
\examples{
\dontrun{
# import some data then export it
dat <- import_isd_hourly(year = 2012)
write_adms(dat, file = "~/adms_met.MET")
}
}
\seealso{
Other Met writing functions:
\code{\link[=write_met]{write_met()}}
}
\concept{ADMS functions}
\concept{Met writing functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_ghcn_monthly.R
\name{import_ghcn_monthly_temp}
\alias{import_ghcn_monthly_temp}
\alias{import_ghcn_monthly_prcp}
\title{Import data from the Global Historical Climatology monthly (GHCNm) database}
\usage{
import_ghcn_monthly_temp(
  table = c("inventory", "data"),
  dataset = c("qcu", "qcf", "qfe")
)

import_ghcn_monthly_prcp(
  station = NULL,
  year = NULL,
  table = c("inventory", "data"),
  progress = rlang::is_interactive()
)
}
\arguments{
\item{table}{Either \code{"inventory"}, \code{"data"}, or both. The tables to read and
return in the output list.}

\item{dataset}{For \code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}}. One of the below options.
More information is available at
\url{https://www.ncei.noaa.gov/pub/data/ghcn/v4/readme.txt}.
\itemize{
\item \code{"qcu"}: Quality Control, Unadjusted
\item \code{"qcf"}: Quality Control, Adjusted, using the Pairwise Homogeneity
Algorithm.
\item \code{"qfe"}: Quality Control, Adjusted, Estimated using the Pairwise
Homogeneity Algorithm. Only the years 1961-2010 are provided. This is to
help maximize station coverage when calculating normals.
}}

\item{station}{For \code{\link[=import_ghcn_monthly_prcp]{import_ghcn_monthly_prcp()}}. The specific stations to
import monthly precipitation data for.}

\item{year}{One or more years of interest. If \code{NULL}, the default, all years
of data available for the chosen \code{station}s will be imported. Note that, in
the GHCNd and GHCNm, files are split by station but not year, so setting a
\code{year} will not speed up the download. Specifying fewer years will improve
the speed of a GHCNh download, however.}

\item{progress}{For \code{\link[=import_ghcn_monthly_prcp]{import_ghcn_monthly_prcp()}}. Show a progress bar when
importing many stations? Defaults to \code{TRUE} in interactive R sessions.
Passed to \code{.progress} in \code{\link[purrr:map]{purrr::map()}}.}
}
\value{
a list of \link[tibble:tibble-package]{tibbles}
}
\description{
This function is a convenient way to access the monthly summaries of the
GHCN. Monthly average temperature is available via
\code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}} and monthly precipitation via
\code{\link[=import_ghcn_monthly_prcp]{import_ghcn_monthly_prcp()}}. Note that these functions can take a few
minutes to run, and parallelism is only enabled for precipitation data.
}
\section{Parallel Processing}{


If you are importing a lot of meteorological data, this can take a long
while. This is because each combination of year and station requires
downloading a separate data file from NOAA's online data directory, and the
time each download takes can quickly add up. Many data import functions in
\code{{worldmet}} can use parallel processing to speed downloading up, powered
by the capable \code{{mirai}} package. If users have any \code{{mirai}} "daemons"
set, these functions will download files in parallel. The greatest benefits
will be seen if you spawn as many daemons as you have cores on your
machine, although one fewer than the available cores is often a good rule
of thumb. Your mileage may vary, however, and naturally spawning more
daemons than station-year combinations will lead to diminishing returns.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set workers - once per session
mirai::daemons(4)

# import lots of data - NB: no change in the import function!
big_met <- import_ghcn_hourly(code = "UKI0000EGLL", year = 2010:2025)
}\if{html}{\out{</div>}}
}

\seealso{
Other GHCN functions:
\code{\link[=import_ghcn_countries]{import_ghcn_countries()}},
\code{\link[=import_ghcn_daily]{import_ghcn_daily()}},
\code{\link[=import_ghcn_hourly]{import_ghcn_hourly()}},
\code{\link[=import_ghcn_inventory]{import_ghcn_inventory()}},
\code{\link[=import_ghcn_stations]{import_ghcn_stations()}}
}
\author{
Jack Davison
}
\concept{GHCN functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_isd_hourly.R
\name{import_isd_hourly}
\alias{import_isd_hourly}
\title{Import Meteorological data from the NOAA Integrated Surface Database (ISD)}
\usage{
import_isd_hourly(
  code = "037720-99999",
  year = 2014,
  hourly = TRUE,
  source = c("delim", "fwf"),
  progress = rlang::is_interactive(),
  quiet = FALSE
)
}
\arguments{
\item{code}{The identifying code as a character string. The code is a
combination of the USAF and the WBAN unique identifiers. The codes are
separated by a \dQuote{-} e.g. \code{code = "037720-99999"}.}

\item{year}{The year to import. This can be a vector of years e.g. \code{year = 2000:2005}.}

\item{hourly}{Should hourly means be calculated? The default is \code{TRUE}. If
\code{FALSE} then the raw data are returned.}

\item{source}{The NOAA ISD service stores files in two formats; as delimited
CSV files (\code{"delim"}) and as fixed width files (\code{"fwf"}).
\code{\link[=import_isd_hourly]{import_isd_hourly()}} defaults to \code{"delim"} but, if the delimited data
store is down, users may wish to try \code{"fwf"} instead. Both data sources
should be identical to one another.}

\item{progress}{Show a progress bar when importing many stations/years?
Defaults to \code{TRUE} in interactive R sessions. Passed to \code{.progress} in
\code{\link[purrr:map]{purrr::map()}} and/or \code{\link[purrr:pmap]{purrr::pmap()}}.}

\item{quiet}{Print missing sites/years to the screen? Defaults to \code{FALSE}.}
}
\value{
Returns a data frame of surface observations. The data frame is
consistent for use with the \code{openair} package. Note that the data are
returned in GMT (UTC) time zone format. Users may wish to express the data
in other time zones, e.g., to merge with air pollution data.
}
\description{
This is the main function to import data from the NOAA Integrated Surface
Database (ISD). The ISD contains detailed surface meteorological data from
around the world for over 30,000 locations.
}
\details{
Note the following units for the main variables:

\describe{

\item{date}{Date/time in POSIXct format. \strong{Note the time zone is GMT (UTC)
and may need to be adjusted to merge with other local data. See details
below.}}

\item{latitude}{Latitude in decimal degrees (-90 to 90).}

\item{longitude}{Longitude in decimal degrees (-180 to 180). Negative numbers
are west of the Greenwich Meridian.}

\item{elevation}{Elevation of site in metres.}

\item{wd}{Wind direction in degrees. 90 is from the east.}

\item{ws}{Wind speed in m/s.}

\item{ceil_hgt}{The height above ground level (AGL) of the lowest cloud or
obscuring phenomena layer aloft with 5/8 or more summation total sky cover,
which may be predominantly opaque, or the vertical visibility into a
surface-based obstruction.}

\item{visibility}{The visibility in metres.}

\item{air_temp}{Air temperature in degrees Celcius.}

\item{dew_point}{The dew point temperature in degrees Celcius.}

\item{atmos_pres}{The sea level pressure in millibars.}

\item{RH}{The relative humidity (\%).}

\item{cl_1,  ...,  cl_3}{Cloud cover for different layers in Oktas (1-8).}

\item{cl}{Maximum of cl_1 to cl_3 cloud cover in Oktas (1-8).}

\item{cl_1_height, ..., cl_3_height}{Height of the cloud base for each later
in metres.}

\item{precip_12}{12-hour precipitation in mm. The sum of this column should
give the annual precipitation.}

\item{precip_6}{6-hour precipitation in mm.}

\item{precip}{This value of precipitation spreads the 12-hour total across
the previous 12 hours.}

\item{pwc}{The description of the present weather description (if
available).}

}

The data are returned in GMT (UTC). It may be necessary to adjust the time
zone when combining with other data. For example, if air quality data were
available for Beijing with time zone set to "Etc/GMT-8" (note the negative
offset even though Beijing is ahead of GMT. See the \code{openair} package and
manual for more details), then the time zone of the met data can be changed
to be the same. One way of doing this would be \code{attr(met$date, "tzone") <- "Etc/GMT-8"} for a meteorological data frame called \code{met}. The two data sets
could then be merged based on \code{date}.
}
\section{Parallel Processing}{


If you are importing a lot of meteorological data, this can take a long
while. This is because each combination of year and station requires
downloading a separate data file from NOAA's online data directory, and the
time each download takes can quickly add up. Many data import functions in
\code{{worldmet}} can use parallel processing to speed downloading up, powered
by the capable \code{{mirai}} package. If users have any \code{{mirai}} "daemons"
set, these functions will download files in parallel. The greatest benefits
will be seen if you spawn as many daemons as you have cores on your
machine, although one fewer than the available cores is often a good rule
of thumb. Your mileage may vary, however, and naturally spawning more
daemons than station-year combinations will lead to diminishing returns.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set workers - once per session
mirai::daemons(4)

# import lots of data - NB: no change in the import function!
big_met <- import_ghcn_hourly(code = "UKI0000EGLL", year = 2010:2025)
}\if{html}{\out{</div>}}
}

\examples{

\dontrun{
# import some data
beijing_met <- import_isd_hourly(code = "545110-99999", year = 2010:2011)

# importing lots of data? use mirai for parallel processing
mirai::daemons(4)
beijing_met2 <- import_isd_hourly(code = "545110-99999", year = 2010:2025)
}
}
\seealso{
Other NOAA ISD functions:
\code{\link[=import_isd_lite]{import_isd_lite()}},
\code{\link[=import_isd_stations]{import_isd_stations()}},
\code{\link[=import_isd_stations_live]{import_isd_stations_live()}}
}
\author{
David Carslaw
}
\concept{NOAA ISD functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{getMeta}
\alias{getMeta}
\alias{getMetaLive}
\alias{importNOAA}
\alias{importNOAAlite}
\title{Deprecated ISD access functions}
\usage{
getMeta(
  site = NULL,
  lat = NULL,
  lon = NULL,
  crs = 4326,
  country = NULL,
  state = NULL,
  n = 10,
  end.year = "current",
  provider = c(`Street Map` = "CartoDB.Voyager", Satellite = "Esri.WorldImagery"),
  plot = TRUE,
  returnMap = FALSE
)

getMetaLive(...)

importNOAA(
  code = "037720-99999",
  year = 2014,
  hourly = TRUE,
  source = c("delim", "fwf"),
  quiet = FALSE,
  path = NA,
  n.cores = NULL
)

importNOAAlite(code = "037720-99999", year = 2025, quiet = FALSE, path = NA)
}
\arguments{
\item{site}{A site name search string e.g. \code{site = "heathrow"}. The search
strings and be partial and can be upper or lower case e.g. \code{site = "HEATHR"}.}

\item{lat, lon, n}{Decimal latitude (\code{lat}) and longitude (\code{lon}) (or other Y/X
coordinate if using a different \code{crs}). If provided, the \code{n} closest ISD
stations to this coordinate will be returned.}

\item{crs}{The coordinate reference system (CRS) of the data, passed to
\code{\link[sf:st_crs]{sf::st_crs()}}. By default this is \href{https://epsg.io/4326}{EPSG:4326}, the
CRS associated with the commonly used latitude and longitude coordinates.
Different coordinate systems can be specified using \code{crs} (e.g., \code{crs = 27700} for the \href{https://epsg.io/27700}{British National Grid}). Note that
non-lat/lng coordinate systems will be re-projected to \code{EPSG:4326} for
making comparisons with the NOAA metadata.}

\item{country}{The country code. This is a two letter code. For a full
listing see \url{https://www.ncei.noaa.gov/pub/data/noaa/isd-history.csv}.}

\item{state}{The state code. This is a two letter code.}

\item{end.year}{To help filter sites based on how recent the available data
are. \code{end.year} can be "current", "any" or a numeric year such as 2016, or
a range of years e.g. 1990:2016 (which would select any site that had an
end date in that range. \strong{By default only sites that have some data for the
current year are returned}.}

\item{provider}{When \code{return = "map"}, by default a map will be created in
which readers may toggle between a vector street map and a satellite/aerial
image. \code{provider} allows users to override this default; see
\url{http://leaflet-extras.github.io/leaflet-providers/preview/} for a list
of all base maps that can be used. Base maps can be toggled using a layer
control menu; the labels will be taken from the name of the base map unless
a named list is defined (see default value).}

\item{plot}{If \code{TRUE} will plot sites on an interactive leaflet map.}

\item{returnMap}{Should the leaflet map be returned instead of the meta data?
Default is \code{FALSE}.}

\item{...}{Currently unused.}

\item{code}{The identifying code as a character string. The code is a
combination of the USAF and the WBAN unique identifiers. The codes are
separated by a \dQuote{-} e.g. \code{code = "037720-99999"}.}

\item{year}{The year to import. This can be a vector of years e.g. \code{year = 2000:2005}.}

\item{hourly}{Should hourly means be calculated? The default is \code{TRUE}. If
\code{FALSE} then the raw data are returned.}

\item{source}{The NOAA ISD service stores files in two formats; as delimited
CSV files (\code{"delim"}) and as fixed width files (\code{"fwf"}).
\code{\link[=import_isd_hourly]{import_isd_hourly()}} defaults to \code{"delim"} but, if the delimited data
store is down, users may wish to try \code{"fwf"} instead. Both data sources
should be identical to one another.}

\item{quiet}{Print missing sites/years to the screen? Defaults to \code{FALSE}.}

\item{path}{If a file path is provided, the data are saved as an rds file at
the chosen location e.g.  \code{path = "C:/Users/David"}. Files are saved by
year and site.}

\item{n.cores}{No longer recommended; please set \code{\link[mirai:daemons]{mirai::daemons()}} in your R
session. This argument is provided for back compatibility, and is passed to
the \code{n} argument of \code{\link[mirai:daemons]{mirai::daemons()}} on behalf of the user. Any set
daemons will be reset once the function completes. Default is \code{NULL}, which
means no parallelism. \code{n.cores = 1L} is equivalent to \code{n.cores = NULL}.}
}
\value{
A data frame is returned with all available meta data, mostly
importantly including a \code{code} that can be supplied to \code{\link[=importNOAA]{importNOAA()}}. If
latitude and longitude searches are made an approximate distance, \code{dist} in
km is also returned.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function is part of an old \code{worldmet} API. Please use the following
alternatives:
\itemize{
\item \code{\link[=getMeta]{getMeta()}} -> \code{\link[=import_isd_stations]{import_isd_stations()}}
\item \code{\link[=getMetaLive]{getMetaLive()}} -> \code{\link[=import_isd_stations_live]{import_isd_stations_live()}}
\item \code{\link[=importNOAA]{importNOAA()}} -> \code{\link[=import_isd_hourly]{import_isd_hourly()}}
\item \code{\link[=importNOAAlite]{importNOAAlite()}} -> \code{\link[=import_isd_lite]{import_isd_lite()}}
\item the \code{path} argument -> \code{\link[=write_met]{write_met()}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deprecated.R
\name{exportADMS}
\alias{exportADMS}
\title{Deprecated data functions}
\usage{
exportADMS(dat, out = "./ADMS_met.MET", interp = FALSE, maxgap = 2)
}
\arguments{
\item{dat}{A data frame imported by \code{\link[=importNOAA]{importNOAA()}}.}

\item{out}{A file name for the ADMS file. The file is written to the working
directory by default.}

\item{interp}{Should interpolation of missing values be undertaken? If \code{TRUE}
linear interpolation is carried out for gaps of up to and including
\code{maxgap}.}

\item{maxgap}{The maximum gap in hours that should be interpolated where
there are missing data when \code{interp = TRUE.} Data with gaps more than
\code{maxgap} are left as missing.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function is part of an old \code{worldmet} API. Please use the following
alternatives:
\itemize{
\item \code{\link[=exportADMS]{exportADMS()}} -> \code{\link[=write_adms]{write_adms()}}
\item the \code{path} argument -> \code{\link[=write_met]{write_met()}}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_ghcn_hourly.R
\name{import_ghcn_hourly}
\alias{import_ghcn_hourly}
\title{Import data from the Global Historical Climatology hourly (GHCNh) database}
\usage{
import_ghcn_hourly(
  station = "UKI0000EGLL",
  year = NULL,
  source = c("psv", "parquet"),
  hourly = TRUE,
  extra = FALSE,
  abbr_names = TRUE,
  append_codes = FALSE,
  codes = c("measurement_code", "quality_code", "report_type", "source_code",
    "source_id"),
  progress = rlang::is_interactive()
)
}
\arguments{
\item{station}{One or more site codes for the station(s) of interest,
obtained using \code{\link[=import_ghcn_stations]{import_ghcn_stations()}}.}

\item{year}{One or more years of interest. If \code{NULL}, the default, all years
of data available for the chosen \code{station}s will be imported. Note that, in
the GHCNd and GHCNm, files are split by station but not year, so setting a
\code{year} will not speed up the download. Specifying fewer years will improve
the speed of a GHCNh download, however.}

\item{source}{There are two identical data formats to read from - \code{"psv"}
(flat, pipe-delimited files) and \code{"parquet"} (a newer, faster, columnar
format). The latter is typically faster, but requires the \code{arrow} package
as an additional dependency. Note that this only applies when \code{year} is not
\code{NULL}; all \code{by-site} files are \code{psv} files.}

\item{hourly}{Should hourly means be calculated? The default is \code{TRUE}. If
\code{FALSE} then the raw data are returned, which can be sub-hourly.}

\item{extra}{Should additional columns be returned? The default, \code{FALSE},
returns an opinionated selection of elements that'll be of most interest to
most users. \code{TRUE} will return everything available.}

\item{abbr_names}{Should column names be abbreviated? When \code{TRUE}, the
default, columns like \code{"wind_direction"} are shortened to \code{"wd"}. When
\code{FALSE}, names will match the raw data, albeit in lower case.}

\item{append_codes}{Logical. Should various codes and flags be appended to
the output dataframe?}

\item{codes}{When \code{append_codes} is \code{TRUE}, which codes should be appended to
the dataframe? Any combination of \code{"measurement_code"}, \code{"quality_code"},
\code{"report_type"}, \code{"source_code"}, and/or \code{"source_id"}.}

\item{progress}{Show a progress bar when importing many stations/years?
Defaults to \code{TRUE} in interactive R sessions. Passed to \code{.progress} in
\code{\link[purrr:map]{purrr::map()}} and/or \code{\link[purrr:pmap]{purrr::pmap()}}.}
}
\value{
a \link[tibble:tibble-package]{tibble}
}
\description{
This function flexibly accesses meteorological data from the GHCNh database.
Users can provide any number of years and stations, and fully control the
sorts of data flag codes that are returned with the data. By default, column
names are shortened for easier use in R, but longer, more descriptive names
can be requested.
}
\section{Data Definition}{


The following core elements are in the GHCNh:
\itemize{
\item \strong{wind_direction:} (\code{wd}) Wind direction from true north (degrees). 360 = north, 180 = south, 270 = west; 000 indicates calm winds.
\item \strong{wind_speed:} (\code{ws}) Average wind speed (m/s).
\item \strong{temperature:} (\code{air_temp}) Air (dry bulb) temperature at approximately 2 meters above ground level, in degrees Celsius (to tenths).
\item \strong{station_level_pressure:} (\code{atmos_pres}) Pressure observed at station elevation; true barometric pressure at the location (hPa).
\item \strong{visibility:} (\code{visibility}) Horizontal visibility distance (km).
\item \strong{dew_point_temperature:} (\code{dew_point}) Dew point temperature in degrees Celsius (to tenths).
\item \strong{relative_humidity:} (\code{rh}) Relative humidity in whole percent, measured or calculated from temperature and dew point.
\item \strong{sky_cover:} (\code{cl}) Maximum of all \strong{sky_cover_X} elements (oktas).
\item \strong{sky_cover_baseht:} (\code{cl_baseht}) The height above ground level (AGL) of the lowest cloud or obscuring phenomena layer aloft with 5/8 or more summation total sky cover, which may be predominantly opaque, or the vertical visibility into a surface-based obstruction.
\item \strong{sky_cover_X:} (\code{cl_X}) Fraction of sky covered by clouds (oktas). Up to 3 layers reported.
\item \strong{sky_cover_baseht_X:} (\code{cl_baseht_X}) Cloud base height for lowest layer (m). Up to 3 layers reported.
\item \strong{precipitation:} (\code{precip}) Total liquid precipitation (rain or melted snow) for the hour (mm). "T" indicates trace amounts.
\item \strong{precipitation_XX_hour:} (\code{precip_XX}) 3-hour total liquid precipitation (mm). "T" indicates trace amounts.
}

When \code{extra = TRUE}, the following additional columns are included:
\itemize{
\item \strong{sea_level_pressure:} (\code{sea_pres}) Estimated pressure at sea level directly below the station using actual temperature profile (hPa).
\item \strong{wind_gust:} (\code{wg}) Peak short-duration wind speed (m/s) exceeding the average wind speed.
\item \strong{wet_bulb_temperature:} (\code{wet_bulb}) Wet bulb temperature in degrees Celsius (to tenths), measured or calculated from temperature, dew point, and station pressure.
\item \strong{snow_depth:} (\code{snow_depth}) Depth of snowpack on the ground (mm).
\item \strong{altimeter:} (\code{altimeter}) Pressure reduced to mean sea level using standard atmosphere profile (hPa).
\item \strong{pressure_3hr_change:} (\code{pres_03}) Change in atmospheric pressure over a 3-hour period (hPa), with tendency code.
}

If \code{hourly = FALSE}, the following character columns may also be present.
\itemize{
\item \strong{pres_wx_MWX:} (\code{wx_mwX}) Present weather observation from manual reports (code). Up to 3 observations per report.
\item \strong{pres_wx_AUX:} (\code{wx_auX}) Present weather observation from automated ASOS/AWOS sensors (code). Up to 3 observations per report.
\item \strong{pres_wx_AWX:} (\code{wx_aqX}) Present weather observation from automated sensors (code). Up to 3 observations per report.
\item \strong{remarks:} (\code{remarks}) Raw observation remarks encoded in ICAO-standard METAR/SYNOP format.
}
}

\section{Parallel Processing}{


If you are importing a lot of meteorological data, this can take a long
while. This is because each combination of year and station requires
downloading a separate data file from NOAA's online data directory, and the
time each download takes can quickly add up. Many data import functions in
\code{{worldmet}} can use parallel processing to speed downloading up, powered
by the capable \code{{mirai}} package. If users have any \code{{mirai}} "daemons"
set, these functions will download files in parallel. The greatest benefits
will be seen if you spawn as many daemons as you have cores on your
machine, although one fewer than the available cores is often a good rule
of thumb. Your mileage may vary, however, and naturally spawning more
daemons than station-year combinations will lead to diminishing returns.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set workers - once per session
mirai::daemons(4)

# import lots of data - NB: no change in the import function!
big_met <- import_ghcn_hourly(code = "UKI0000EGLL", year = 2010:2025)
}\if{html}{\out{</div>}}
}

\seealso{
Other GHCN functions:
\code{\link[=import_ghcn_countries]{import_ghcn_countries()}},
\code{\link[=import_ghcn_daily]{import_ghcn_daily()}},
\code{\link[=import_ghcn_inventory]{import_ghcn_inventory()}},
\code{\link[=import_ghcn_monthly_temp]{import_ghcn_monthly_temp()}},
\code{\link[=import_ghcn_stations]{import_ghcn_stations()}}
}
\author{
Jack Davison
}
\concept{GHCN functions}
